#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import io
import sys

from ax.utils.common.base import Base
from ax.utils.common.testutils import TestCase


# pyre-fixme[3]: Return type must be annotated.
def _f():
    e = RuntimeError("Test")
    raise e


F_FAILURE_LINENO = 19  # Line # for the error in `_f`.


def _g() -> None:
    _f()  # Lines along the path are matched too


class MyBase(Base):
    def __init__(self, val: str) -> None:
        self.field = val


class TestTestUtils(TestCase):
    def test_equal(self) -> None:
        try:  # Check case where values aren't `Base` subclasses.
            self.assertEqual(1, 2)
        except AssertionError as err:
            self.assertEqual(str(err), "1 != 2")

        try:  # Check case where values are `Base` subclasses.
            self.assertEqual(MyBase("red"), MyBase("panda"))
        except AssertionError as err:
            expected_suffix = (
                "\n\nFields with different values:\n\n1) field: red "
                "(type <class 'str'>) != panda (type <class 'str'>)"
            )
            self.assertIn(expected_suffix, str(err))

    def test_raises_on(self) -> None:
        with self.assertRaisesOn(RuntimeError, "raise e"):
            _f()

        # Check that we fail if the source line is not what we expect
        with self.assertRaisesRegex(Exception, "was not found in the traceback"):
            with self.assertRaisesOn(RuntimeError, 'raise Exception("Test")'):
                _f()

        # Check that the exception passes through if it's not the one we meant to catch
        with self.assertRaisesRegex(RuntimeError, "Test"):
            with self.assertRaisesOn(AssertionError, "raise e"):
                _f()

        with self.assertRaisesOn(
            RuntimeError, "_f()  # Lines along the path are matched too"
        ):
            _g()

        # Use this as a context manager to get the position of an error
        with self.assertRaisesOn(RuntimeError) as cm:
            _f()
        # pyre-fixme[16]: `None` has no attribute `filename`.
        self.assertEqual(cm.filename, __file__)
        # pyre-fixme[16]: `None` has no attribute `lineno`.
        self.assertEqual(cm.lineno, F_FAILURE_LINENO)

    def test_silence_warning_normal(self) -> None:
        new_stderr = io.StringIO()
        old_err = sys.stderr
        try:
            sys.stderr = new_stderr
            with self.silence_stderr():
                print("A message", file=sys.stderr)
        finally:
            sys.stderr = old_err
        self.assertEqual(new_stderr.getvalue(), "")

    def test_silence_warning(self) -> None:
        new_stderr = io.StringIO()
        old_err = sys.stderr
        with self.assertRaises(AssertionError):
            try:
                sys.stderr = new_stderr
                with self.silence_stderr():
                    print("A message", file=sys.stderr)
                    raise AssertionError()
            finally:
                sys.stderr = old_err
        self.assertTrue(new_stderr.getvalue().startswith("A message\n"))

    def test_fail_deprecated(self) -> None:
        self.assertEqual(1, 1)
        with self.assertRaises(RuntimeError):
            self.assertEquals(1, 1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from dataclasses import dataclass
from typing import Dict, Iterable, List, Optional, Sequence, Union
from unittest import mock

import numpy as np
import pandas as pd
from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial
from ax.core.experiment import Experiment
from ax.core.map_data import MapData
from ax.core.objective import Objective
from ax.core.optimization_config import OptimizationConfig
from ax.metrics.tensorboard import TensorboardCurveMetric, TensorboardMetric
from ax.runners.synthetic import SyntheticRunner
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_search_space, get_trial
from pyre_extensions import assert_is_instance
from tensorboard.backend.event_processing import event_multiplexer


@dataclass
class _TensorProto:
    double_val: List[float]


@dataclass
class _TensorEvent:
    step: int
    tensor_proto: _TensorProto


def _get_fake_multiplexer(
    fake_data: Sequence[float],
) -> event_multiplexer.EventMultiplexer:
    mul = event_multiplexer.EventMultiplexer()

    # pyre-ignore[8] Return fake tags when content is requested
    mul.PluginRunToTagToContent = lambda plugin: {".": {"loss": ""}}

    # pyre-ignore[8] Return fake data when tensors requested
    mul.Tensors = lambda run, tag: [
        _TensorEvent(step=i, tensor_proto=_TensorProto(double_val=[dat]))
        for i, dat in enumerate(fake_data)
    ]

    return mul


class TensorboardMetricTest(TestCase):
    def test_fetch_trial_data(self) -> None:
        fake_data = [8.0, 9.0, 2.0, 1.0]
        fake_multiplexer = _get_fake_multiplexer(fake_data=fake_data)

        with mock.patch.object(
            TensorboardMetric,
            "_get_event_multiplexer_for_trial",
            return_value=fake_multiplexer,
        ):
            metric = TensorboardMetric(
                name="loss", tag="loss", lower_is_better=True, smoothing=0
            )
            trial = get_trial()

            result = metric.fetch_trial_data(trial=trial)

            df = assert_is_instance(result.unwrap(), MapData).map_df

            expected_df = pd.DataFrame(
                [
                    {
                        "arm_name": "0_0",
                        "metric_name": "loss",
                        "mean": fake_data[i],
                        "sem": float("nan"),
                        "trial_index": 0,
                        "step": float(i),
                    }
                    for i in range(len(fake_data))
                ]
            )

            self.assertTrue(df.equals(expected_df))

    def test_smoothing(self) -> None:
        fake_data = [8.0, 4.0, 2.0, 1.0]
        smoothing = 0.5
        smooth_data = pd.Series(fake_data).ewm(alpha=smoothing).mean().tolist()

        fake_multiplexer = _get_fake_multiplexer(fake_data=fake_data)

        with mock.patch.object(
            TensorboardMetric,
            "_get_event_multiplexer_for_trial",
            return_value=fake_multiplexer,
        ):
            metric = TensorboardMetric(
                name="loss", tag="loss", lower_is_better=True, smoothing=smoothing
            )
            trial = get_trial()

            result = metric.fetch_trial_data(trial=trial)

            df = assert_is_instance(result.unwrap(), MapData).map_df

            expected_df = pd.DataFrame(
                [
                    {
                        "arm_name": "0_0",
                        "metric_name": "loss",
                        "mean": smooth_data[i],
                        "sem": float("nan"),
                        "trial_index": 0,
                        "step": float(i),
                    }
                    for i in range(len(fake_data))
                ]
            )

            self.assertTrue(df.equals(expected_df))

    def test_cumulative_best(self) -> None:
        fake_data = [4.0, 8.0, 2.0, 1.0]
        cummin_data = pd.Series(fake_data).cummin().tolist()

        fake_multiplexer = _get_fake_multiplexer(fake_data=fake_data)

        with mock.patch.object(
            TensorboardMetric,
            "_get_event_multiplexer_for_trial",
            return_value=fake_multiplexer,
        ):
            metric = TensorboardMetric(
                name="loss",
                tag="loss",
                lower_is_better=True,
                cumulative_best=True,
                smoothing=0,
            )
            trial = get_trial()

            result = metric.fetch_trial_data(trial=trial)

            df = assert_is_instance(result.unwrap(), MapData).map_df

            expected_df = pd.DataFrame(
                [
                    {
                        "arm_name": "0_0",
                        "metric_name": "loss",
                        "mean": cummin_data[i],
                        "sem": float("nan"),
                        "trial_index": 0,
                        "step": float(i),
                    }
                    for i in range(len(fake_data))
                ]
            )

            self.assertTrue(df.equals(expected_df))


class TensorboardCurveMetricTest(TestCase):
    def test_tensorboard_curve_metric(self) -> None:
        def mock_get_tb_from_posix(
            path: str, tags: Optional[List[str]] = None
        ) -> Dict[str, pd.Series]:
            data = np.array([10, 3, 5, 2, 7, 1])
            return {"test_curve": pd.Series((int(path) + 1) * data)}

        mock_path = "ax.metrics.tensorboard.get_tb_from_posix"

        class FakeTensorboardCurveMetric(TensorboardCurveMetric):
            @classmethod
            def get_ids_from_trials(
                cls, trials: Iterable[BaseTrial]
            ) -> Dict[int, Union[int, str]]:
                result = {}
                for trial in trials:
                    result[trial.index] = trial.index
                return result

        with mock.patch(mock_path, side_effect=mock_get_tb_from_posix):
            # test simple
            experiment = Experiment(
                name="dummy_experiment",
                search_space=get_branin_search_space(),
                optimization_config=OptimizationConfig(
                    objective=Objective(
                        metric=FakeTensorboardCurveMetric(
                            name="test_metric",
                            curve_name="test_curve",
                            lower_is_better=True,
                            cumulative_best=False,
                        ),
                        minimize=True,
                    )
                ),
                runner=SyntheticRunner(),
            )
            for param in range(0, 2):
                trial = experiment.new_trial()
                trial.add_arm(Arm(parameters={"x1": float(param), "x2": 0.0}))
                trial.run()

            self.assertTrue(
                np.allclose(
                    # pyre-fixme[16]: `Data` has no attribute `map_df`.
                    experiment.fetch_data().map_df["mean"].to_numpy(),
                    np.array(
                        [10.0, 3.0, 5.0, 2.0, 7.0, 1.0, 20.0, 6.0, 10.0, 4.0, 14.0, 2.0]
                    ),
                )
            )

            # test cumulative best
            experiment = Experiment(
                name="dummy_experiment",
                search_space=get_branin_search_space(),
                optimization_config=OptimizationConfig(
                    objective=Objective(
                        metric=FakeTensorboardCurveMetric(
                            name="test_metric",
                            curve_name="test_curve",
                            lower_is_better=True,
                            cumulative_best=True,
                        ),
                        minimize=True,
                    )
                ),
                runner=SyntheticRunner(),
            )
            for param in range(0, 2):
                trial = experiment.new_trial()
                trial.add_arm(Arm(parameters={"x1": float(param), "x2": 0.0}))
                trial.run()

            self.assertTrue(
                np.allclose(
                    experiment.fetch_data().map_df["mean"].to_numpy(),
                    np.array(
                        [10.0, 3.0, 3.0, 2.0, 2.0, 1.0, 20.0, 6.0, 6.0, 4.0, 4.0, 2.0]
                    ),
                )
            )

            # test cumulative best (lower is worse)
            experiment = Experiment(
                name="dummy_experiment",
                search_space=get_branin_search_space(),
                optimization_config=OptimizationConfig(
                    objective=Objective(
                        metric=FakeTensorboardCurveMetric(
                            name="test_metric",
                            curve_name="test_curve",
                            lower_is_better=False,
                            cumulative_best=True,
                        ),
                        minimize=False,
                    )
                ),
                runner=SyntheticRunner(),
            )
            for param in range(0, 2):
                trial = experiment.new_trial()
                trial.add_arm(Arm(parameters={"x1": float(param), "x2": 0.0}))
                trial.run()

            self.assertTrue(
                np.allclose(
                    experiment.fetch_data().map_df["mean"].to_numpy(),
                    np.array(
                        [
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            20.0,
                            20.0,
                            20.0,
                            20.0,
                            20.0,
                            20.0,
                        ]
                    ),
                ),
            )

            # test smoothing
            experiment = Experiment(
                name="dummy_experiment",
                search_space=get_branin_search_space(),
                optimization_config=OptimizationConfig(
                    objective=Objective(
                        metric=FakeTensorboardCurveMetric(
                            name="test_metric",
                            curve_name="test_curve",
                            lower_is_better=True,
                            cumulative_best=False,
                            smoothing_window=3,
                        ),
                        minimize=True,
                    )
                ),
                runner=SyntheticRunner(),
            )
            for param in range(0, 2):
                trial = experiment.new_trial()
                trial.add_arm(Arm(parameters={"x1": float(param), "x2": 0.0}))
                trial.run()
            self.assertTrue(
                np.allclose(
                    experiment.fetch_data().map_df["mean"].to_numpy(),
                    np.array(
                        [
                            6.00000000,
                            6.00000000,
                            6.00000000,
                            3.33333333,
                            4.66666667,
                            3.33333333,
                            12.0,
                            12.0,
                            12.0,
                            6.66666667,
                            9.33333333,
                            6.66666667,
                        ]
                    ),
                )
            )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from typing import DefaultDict, Dict, List, Optional

import numpy as np
from ax.core.arm import Arm
from ax.core.observation import Observation, ObservationData, separate_observations
from ax.core.search_space import SearchSpace
from ax.modelbridge.base import ModelBridge
from ax.modelbridge.transforms.base import Transform
from ax.models.types import TConfig


class MergeRepeatedMeasurements(Transform):
    """Merge repeated measurements for to obtain one observation per arm.

    Repeated measurements are merged via inverse variance weighting (e.g. over
    different trials). This intentionally ignores the trial index and assumes
    stationarity.

    TODO: Support inverse variance weighting correlated outcomes (full covariance).

    Note: this is not reversible.
    """

    def __init__(
        self,
        search_space: Optional[SearchSpace] = None,
        observations: Optional[List[Observation]] = None,
        modelbridge: Optional[ModelBridge] = None,
        config: Optional[TConfig] = None,
    ) -> None:
        if observations is None:
            raise RuntimeError("MergeRepeatedMeasurements requires observations")
        # create a mapping of arm_key -> {metric_name: {means: [], vars: []}}
        arm_to_multi_obs: DefaultDict[
            str, DefaultDict[str, DefaultDict[str, List[float]]]
        ] = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        observation_features, observation_data = separate_observations(observations)
        for j, obsd in enumerate(observation_data):
            # This intentionally ignores the trial index
            key = Arm.md5hash(observation_features[j].parameters)
            # TODO: support inverse variance weighting for multivariate distributions
            # (full covariance)
            diag = np.diag(np.diag(obsd.covariance))
            if np.any(np.isnan(obsd.covariance)):
                raise NotImplementedError("All metrics must have noise observations.")
            elif ~np.all(obsd.covariance == diag):
                raise NotImplementedError(
                    "Only independent metrics are currently supported."
                )
            for i, m in enumerate(obsd.metric_names):
                arm_to_multi_obs[key][m]["means"].append(obsd.means[i])
                arm_to_multi_obs[key][m]["vars"].append(obsd.covariance[i, i])

        self.arm_to_merged: DefaultDict[str, Dict[str, Dict[str, float]]] = defaultdict(
            dict
        )
        for k, metric_dict in arm_to_multi_obs.items():
            for m, v in metric_dict.items():
                # inverse variance weighting
                var = np.array(v["vars"])
                means = np.array(v["means"])
                noiseless = var == 0
                if np.any(noiseless):
                    noiseless_means = means[noiseless]
                    if (noiseless_means.shape[0] > 1) and (
                        not np.all(noiseless_means[1:] == noiseless_means[0])
                    ):
                        raise ValueError(
                            "All repeated arms with noiseless measurements "
                            "must have the same means."
                        )
                    self.arm_to_merged[k][m] = {
                        "mean": noiseless_means[0],
                        "var": 0.0,
                    }
                else:
                    inv_var = 1 / np.array(var)
                    inv_sum_inv_var = 1 / np.sum(inv_var)
                    weights = inv_var * inv_sum_inv_var
                    self.arm_to_merged[k][m] = {
                        "mean": np.sum(means * weights),
                        "var": inv_sum_inv_var,
                    }

    def transform_observations(
        self,
        observations: List[Observation],
    ) -> List[Observation]:
        # Transform observations
        new_observations = []
        observation_features, observation_data = separate_observations(observations)
        arm_to_merged = deepcopy(self.arm_to_merged)
        for j, obsd in enumerate(observation_data):
            key = Arm.md5hash(observation_features[j].parameters)
            # pop to ensure that the resulting observations list has one
            # observation per unique arm
            metric_dict = arm_to_merged.pop(key, None)
            if metric_dict is None:
                continue
            merged_means = np.zeros(len(obsd.metric_names))
            merged_covariance = np.zeros(
                (len(obsd.metric_names), len(obsd.metric_names))
            )
            for i, m in enumerate(obsd.metric_names):
                merged_metric = metric_dict[m]
                merged_means[i] = merged_metric["mean"]
                merged_covariance[i, i] = merged_metric["var"]
            new_obsd = ObservationData(
                metric_names=obsd.metric_names,
                means=merged_means,
                covariance=merged_covariance,
            )
            new_obs = Observation(
                features=observation_features[j],
                data=new_obsd,
                arm_name=observations[j].arm_name,
            )
            new_observations.append(new_obs)
        return new_observations



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import List, Optional

from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generation_strategy_interface import GenerationStrategyInterface
from ax.core.generator_run import GeneratorRun
from ax.exceptions.core import AxError, UnsupportedError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_experiment, SpecialGenerationStrategy


class MyGSI(GenerationStrategyInterface):
    def gen_for_multiple_trials_with_multiple_models(
        self,
        experiment: Experiment,
        data: Optional[Data] = None,
        # TODO[drfreund, danielcohennyc, mgarrard]: Update the format of the arguments
        # below as we find the right one.
        num_generator_runs: int = 1,
        n: int = 1,
    ) -> List[List[GeneratorRun]]:
        raise NotImplementedError

    def clone_reset(self) -> "MyGSI":
        raise NotImplementedError


class TestGenerationStrategyInterface(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.exp = get_experiment()
        self.gsi = MyGSI(name="my_GSI")
        self.special_gsi = SpecialGenerationStrategy()

    def test_constructor(self) -> None:
        with self.assertRaisesRegex(TypeError, ".* abstract"):
            GenerationStrategyInterface(name="my_GSI")  # pyre-ignore[45]
        self.assertEqual(self.gsi.name, "my_GSI")

    def test_abstract(self) -> None:
        with self.assertRaises(NotImplementedError):
            self.gsi.gen_for_multiple_trials_with_multiple_models(experiment=self.exp)

        with self.assertRaises(NotImplementedError):
            self.gsi.clone_reset()

    def test_experiment(self) -> None:
        with self.assertRaisesRegex(AxError, "No experiment"):
            self.gsi.experiment
        self.gsi.experiment = self.exp
        exp_2 = get_experiment()
        exp_2.name = "exp_2"
        with self.assertRaisesRegex(UnsupportedError, "has been used for"):
            self.gsi.experiment = exp_2



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import json
import logging
import warnings
from functools import partial

from logging import Logger
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import ax.service.utils.early_stopping as early_stopping_utils
import numpy as np
import pandas as pd
import torch
from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.experiment import DataType, Experiment
from ax.core.formatting_utils import data_and_evaluations_from_raw_data
from ax.core.generation_strategy_interface import GenerationStrategyInterface
from ax.core.generator_run import GeneratorRun
from ax.core.map_data import MapData
from ax.core.map_metric import MapMetric
from ax.core.objective import MultiObjective, Objective
from ax.core.observation import ObservationFeatures
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.trial import Trial
from ax.core.types import (
    TEvaluationOutcome,
    TModelPredictArm,
    TParameterization,
    TParamValue,
)
from ax.core.utils import get_pending_observation_features_based_on_trial_status
from ax.early_stopping.strategies import BaseEarlyStoppingStrategy
from ax.early_stopping.utils import estimate_early_stopping_savings
from ax.exceptions.constants import CHOLESKY_ERROR_ANNOTATION
from ax.exceptions.core import (
    DataRequiredError,
    OptimizationComplete,
    OptimizationShouldStop,
    UnsupportedError,
    UnsupportedPlotError,
)
from ax.exceptions.generation_strategy import MaxParallelismReachedException
from ax.global_stopping.strategies.base import BaseGlobalStoppingStrategy
from ax.global_stopping.strategies.improvement import constraint_satisfaction
from ax.modelbridge.dispatch_utils import choose_generation_strategy
from ax.modelbridge.generation_strategy import GenerationStrategy
from ax.modelbridge.prediction_utils import predict_by_features
from ax.plot.base import AxPlotConfig
from ax.plot.contour import plot_contour
from ax.plot.feature_importances import plot_feature_importance_by_feature
from ax.plot.helper import _format_dict
from ax.plot.trace import optimization_trace_single_method
from ax.service.utils.best_point_mixin import BestPointMixin
from ax.service.utils.instantiation import (
    FixedFeatures,
    InstantiationBase,
    ObjectiveProperties,
)
from ax.service.utils.report_utils import exp_to_df
from ax.service.utils.with_db_settings_base import DBSettings, WithDBSettingsBase
from ax.storage.json_store.decoder import (
    generation_strategy_from_json,
    object_from_json,
)
from ax.storage.json_store.encoder import object_to_json
from ax.storage.json_store.registry import (
    CORE_CLASS_DECODER_REGISTRY,
    CORE_CLASS_ENCODER_REGISTRY,
    CORE_DECODER_REGISTRY,
    CORE_ENCODER_REGISTRY,
    TDecoderRegistry,
)
from ax.utils.common.docutils import copy_doc
from ax.utils.common.executils import retry_on_exception
from ax.utils.common.logger import _round_floats_for_logging, get_logger
from ax.utils.common.random import with_rng_seed
from ax.utils.common.typeutils import checked_cast, not_none
from pyre_extensions import assert_is_instance

logger: Logger = get_logger(__name__)


AxClientSubclass = TypeVar("AxClientSubclass", bound="AxClient")

ROUND_FLOATS_IN_LOGS_TO_DECIMAL_PLACES: int = 6

# pyre-fixme[5]: Global expression must be annotated.
round_floats_for_logging = partial(
    _round_floats_for_logging,
    decimal_places=ROUND_FLOATS_IN_LOGS_TO_DECIMAL_PLACES,
)


class AxClient(WithDBSettingsBase, BestPointMixin, InstantiationBase):
    """
    Convenience handler for management of experimentation cycle through a
    service-like API. External system manages scheduling of the cycle and makes
    calls to this client to get next suggestion in the experiment and log back
    data from the evaluation of that suggestion.

    Note: `AxClient` expects to only propose 1 arm (suggestion) per trial; support
    for use cases that require use of batches is coming soon.

    Two custom types used in this class for convenience are `TParamValue` and
    `TParameterization`. Those are shortcuts for `Union[str, bool, float, int]`
    and `Dict[str, Union[str, bool, float, int]]`, respectively.

    Args:
        generation_strategy: Optional generation strategy. If not set, one is
            intelligently chosen based on properties of search space.

        db_settings: Settings for saving and reloading the underlying experiment
            to a database. Expected to be of type
            ax.storage.sqa_store.structs.DBSettings and require SQLAlchemy.

        enforce_sequential_optimization: Whether to enforce that when it is
            reasonable to switch models during the optimization (as prescribed
            by `num_trials` in generation strategy), Ax will wait for enough trials
            to be completed with data to proceed. Defaults to True. If set to
            False, Ax will keep generating new trials from the previous model
            until enough data is gathered. Use this only if necessary;
            otherwise, it is more resource-efficient to
            optimize sequentially, by waiting until enough data is available to
            use the next model.

        random_seed: Optional integer random seed, set to fix the optimization
            random seed for reproducibility. Works only for Sobol quasi-random
            generator and for BoTorch-powered models. For the latter models, the
            trials generated from the same optimization setup with the same seed,
            will be mostly similar, but the exact parameter values may still vary
            and trials latter in the optimizations will diverge more and more.
            This is because a degree of randomness is essential for high performance
            of the Bayesian optimization models and is not controlled by the seed.

            Note: In multi-threaded environments, the random seed is thread-safe,
            but does not actually guarantee reproducibility. Whether the outcomes
            will be exactly the same for two same operations that use the random
            seed, depends on whether the threads modify the random state in the
            same order across the two operations.

        torch_device: An optional `torch.device` object, used to choose the device
            used for generating new points for trials. Works only for torch-based
            models, such as GPEI. Ignored if a `generation_strategy` is passed in
            manually. To specify the device for a custom `generation_strategy`,
            pass in `torch_device` as part of `model_kwargs`. See
            https://ax.dev/tutorials/generation_strategy.html for a tutorial on
            generation strategies.

        verbose_logging: Whether Ax should log significant optimization events,
            defaults to `True`.

        suppress_storage_errors: Whether to suppress SQL storage-related errors if
            encountered. Only use if SQL storage is not important for the given use
            case, since this will only log, but not raise, an exception if its
            encountered while saving to DB or loading from it.

        early_stopping_strategy: A ``BaseEarlyStoppingStrategy`` that determines
            whether a trial should be stopped given the current state of
            the experiment. Used in ``should_stop_trials_early``.

        global_stopping_strategy: A ``BaseGlobalStoppingStrategy`` that determines
            whether the full optimization should be stopped or not.
    """

    _experiment: Optional[Experiment] = None

    def __init__(
        self,
        generation_strategy: Optional[GenerationStrategy] = None,
        db_settings: Optional[DBSettings] = None,
        enforce_sequential_optimization: bool = True,
        random_seed: Optional[int] = None,
        torch_device: Optional[torch.device] = None,
        verbose_logging: bool = True,
        suppress_storage_errors: bool = False,
        early_stopping_strategy: Optional[BaseEarlyStoppingStrategy] = None,
        global_stopping_strategy: Optional[BaseGlobalStoppingStrategy] = None,
    ) -> None:
        super().__init__(
            db_settings=db_settings,
            suppress_all_errors=suppress_storage_errors,
        )

        if not verbose_logging:
            logger.setLevel(logging.WARNING)
        else:
            logger.info(
                "Starting optimization with verbose logging. To disable logging, "
                "set the `verbose_logging` argument to `False`. Note that float "
                "values in the logs are rounded to "
                f"{ROUND_FLOATS_IN_LOGS_TO_DECIMAL_PLACES} decimal points."
            )
        if generation_strategy is not None and torch_device is not None:
            warnings.warn(
                "Both a `generation_strategy` and a `torch_device` were specified. "
                "`torch_device` will be ignored. Instead, specify `torch_device` "
                "by passing it in `model_kwargs` while creating the "
                "`generation_strategy`.",
                RuntimeWarning,
                stacklevel=2,
            )
        self._generation_strategy = generation_strategy
        self._enforce_sequential_optimization = enforce_sequential_optimization
        self._random_seed = random_seed
        self._torch_device = torch_device
        self._suppress_storage_errors = suppress_storage_errors
        self._early_stopping_strategy = early_stopping_strategy
        self._global_stopping_strategy = global_stopping_strategy
        if random_seed is not None:
            logger.warning(
                f"Random seed set to {random_seed}. Note that this setting "
                "only affects the Sobol quasi-random generator "
                "and BoTorch-powered Bayesian optimization models. For the latter "
                "models, setting random seed to the same number for two optimizations "
                "will make the generated trials similar, but not exactly the same, "
                "and over time the trials will diverge more."
            )

    # ------------------------ Public API methods. ------------------------

    def create_experiment(
        self,
        parameters: List[
            Dict[str, Union[TParamValue, Sequence[TParamValue], Dict[str, List[str]]]]
        ],
        name: Optional[str] = None,
        description: Optional[str] = None,
        owners: Optional[List[str]] = None,
        objectives: Optional[Dict[str, ObjectiveProperties]] = None,
        parameter_constraints: Optional[List[str]] = None,
        outcome_constraints: Optional[List[str]] = None,
        status_quo: Optional[TParameterization] = None,
        overwrite_existing_experiment: bool = False,
        experiment_type: Optional[str] = None,
        tracking_metric_names: Optional[List[str]] = None,
        choose_generation_strategy_kwargs: Optional[Dict[str, Any]] = None,
        support_intermediate_data: bool = False,
        immutable_search_space_and_opt_config: bool = True,
        is_test: bool = False,
        metric_definitions: Optional[Dict[str, Dict[str, Any]]] = None,
    ) -> None:
        """Create a new experiment and save it if DBSettings available.

        Args:
            parameters: List of dictionaries representing parameters in the
                experiment search space.
                Required elements in the dictionaries are:
                1. "name" (name of parameter, string),
                2. "type" (type of parameter: "range", "fixed", or "choice", string),
                and one of the following:
                3a. "bounds" for range parameters (list of two values, lower bound
                first),
                3b. "values" for choice parameters (list of values), or
                3c. "value" for fixed parameters (single value).
                Optional elements are:
                1. "log_scale" (for float-valued range parameters, bool),
                2. "value_type" (to specify type that values of this parameter should
                take; expects "float", "int", "bool" or "str"),
                3. "is_fidelity" (bool) and "target_value" (float) for fidelity
                parameters,
                4. "is_ordered" (bool) for choice parameters, and
                5. "is_task" (bool) for task parameters.
                6. "digits" (int) for float-valued range parameters.
            name: Name of the experiment to be created.
            description: Description of the experiment to be created.
            objectives: Mapping from an objective name to object containing:
                minimize: Whether this experiment represents a minimization problem.
                threshold: The bound in the objective's threshold constraint.
            parameter_constraints: List of string representation of parameter
                constraints, such as "x3 >= x4" or "-x3 + 2*x4 - 3.5*x5 >= 2". For
                the latter constraints, any number of arguments is accepted, and
                acceptable operators are "<=" and ">=". Note that parameter
                constraints may only be placed on range parameters, not choice
                parameters or fixed parameters.
            outcome_constraints: List of string representation of outcome
                constraints of form "metric_name >= bound", like "m1 <= 3."
            status_quo: Parameterization of the current state of the system.
                If set, this will be added to each trial to be evaluated alongside
                test configurations.
            overwrite_existing_experiment: If an experiment has already been set
                on this `AxClient` instance, whether to reset it to the new one.
                If overwriting the experiment, generation strategy will be
                re-selected for the new experiment and restarted.
                To protect experiments in production, one cannot overwrite existing
                experiments if the experiment is already stored in the database,
                regardless of the value of `overwrite_existing_experiment`.
            tracking_metric_names: Names of additional tracking metrics not used for
                optimization.
            choose_generation_strategy_kwargs: Keyword arguments to pass to
                `choose_generation_strategy` function which determines what
                generation strategy should be used when none was specified on init.
            support_intermediate_data: Whether trials may report intermediate results
                for trials that are still running (i.e. have not been completed via
                `ax_client.complete_trial`).
            immutable_search_space_and_opt_config: Whether it's possible to update the
                search space and optimization config on this experiment after creation.
                Defaults to True. If set to True, we won't store or load copies of the
                search space and optimization config on each generator run, which will
                improve storage performance.
            is_test: Whether this experiment will be a test experiment (useful for
                marking test experiments in storage etc). Defaults to False.
            metric_definitions: A mapping of metric names to extra kwargs to pass
                to that metric. Note these are modified in-place. Each
                Metric must have its own dictionary (metrics cannot share a
                single dictionary object).
        """
        self._validate_early_stopping_strategy(support_intermediate_data)

        objective_kwargs = {}
        if objectives is not None:
            objective_kwargs["objectives"] = {
                objective: ("minimize" if properties.minimize else "maximize")
                for objective, properties in objectives.items()
            }
            if len(objectives.keys()) > 1:
                objective_kwargs["objective_thresholds"] = (
                    self.build_objective_thresholds(objectives)
                )

        experiment = self.make_experiment(
            name=name,
            description=description,
            owners=owners,
            parameters=parameters,
            parameter_constraints=parameter_constraints,
            outcome_constraints=outcome_constraints,
            status_quo=status_quo,
            experiment_type=experiment_type,
            tracking_metric_names=tracking_metric_names,
            metric_definitions=metric_definitions,
            support_intermediate_data=support_intermediate_data,
            immutable_search_space_and_opt_config=immutable_search_space_and_opt_config,
            is_test=is_test,
            **objective_kwargs,
        )
        self._set_runner(experiment=experiment)
        self._set_experiment(
            experiment=experiment,
            overwrite_existing_experiment=overwrite_existing_experiment,
        )
        self._set_generation_strategy(
            choose_generation_strategy_kwargs=choose_generation_strategy_kwargs
        )
        self._save_generation_strategy_to_db_if_possible()

    @property
    def status_quo(self) -> Optional[TParameterization]:
        """The parameterization of the status quo arm of the experiment."""
        if self.experiment.status_quo:
            return self.experiment.status_quo.parameters
        return None

    def set_status_quo(self, params: Optional[TParameterization]) -> None:
        """Set, or unset status quo on the experiment.  There may be risk
        in using this after a trial with the status quo arm has run.

        Args:
            status_quo: Parameterization of the current state of the system.
                If set, this will be added to each trial to be evaluated alongside
                test configurations.
        """
        self.experiment.status_quo = None if params is None else Arm(parameters=params)

    def set_optimization_config(
        self,
        objectives: Optional[Dict[str, ObjectiveProperties]] = None,
        outcome_constraints: Optional[List[str]] = None,
        metric_definitions: Optional[Dict[str, Dict[str, Any]]] = None,
    ) -> None:
        """Overwrite experiment's optimization config

        Args:
            objectives: Mapping from an objective name to object containing:
                minimize: Whether this experiment represents a minimization problem.
                threshold: The bound in the objective's threshold constraint.
            outcome_constraints: List of string representation of outcome
                constraints of form "metric_name >= bound", like "m1 <= 3."
            metric_definitions: A mapping of metric names to extra kwargs to pass
                to that metric
        """
        optimization_config = self.make_optimization_config_from_properties(
            objectives=objectives,
            outcome_constraints=outcome_constraints,
            status_quo_defined=self.experiment.status_quo is not None,
            metric_definitions=metric_definitions,
        )
        if optimization_config:
            self.experiment.optimization_config = optimization_config
            self._save_experiment_to_db_if_possible(
                experiment=self.experiment,
            )
        else:
            raise ValueError(
                "optimization config not set because it was missing objectives"
            )

    def add_tracking_metrics(
        self,
        metric_names: List[str],
        metric_definitions: Optional[Dict[str, Dict[str, Any]]] = None,
    ) -> None:
        """Add a list of new metrics to the experiment.

        If any of the metrics are already defined on the experiment,
        we raise an error and don't add any of them to the experiment

        Args:
            metric_names: Names of metrics to be added.
            metric_definitions: A mapping of metric names to extra kwargs to pass
                to that metric. Note these are modified in-place. Each
                Metric must have its is own dictionary (metrics cannot share a
                single dictionary object).
        """
        self.experiment.add_tracking_metrics(
            metrics=[
                self._make_metric(
                    name=metric_name, metric_definitions=metric_definitions
                )
                for metric_name in metric_names
            ]
        )

    @copy_doc(Experiment.remove_tracking_metric)
    def remove_tracking_metric(self, metric_name: str) -> None:
        self.experiment.remove_tracking_metric(metric_name=metric_name)

    def set_search_space(
        self,
        parameters: List[
            Dict[str, Union[TParamValue, Sequence[TParamValue], Dict[str, List[str]]]]
        ],
        parameter_constraints: Optional[List[str]] = None,
    ) -> None:
        """Sets the search space on the experiment and saves.
        This is expected to fail on base AxClient as experiment will have
        immutable search space and optimization config set to True by default

        Args:
            parameters: List of dictionaries representing parameters in the
                experiment search space.
                Required elements in the dictionaries are:
                1. "name" (name of parameter, string),
                2. "type" (type of parameter: "range", "fixed", or "choice", string),
                and one of the following:
                3a. "bounds" for range parameters (list of two values, lower bound
                first),
                3b. "values" for choice parameters (list of values), or
                3c. "value" for fixed parameters (single value).
                Optional elements are:
                1. "log_scale" (for float-valued range parameters, bool),
                2. "value_type" (to specify type that values of this parameter should
                take; expects "float", "int", "bool" or "str"),
                3. "is_fidelity" (bool) and "target_value" (float) for fidelity
                parameters,
                4. "is_ordered" (bool) for choice parameters, and
                5. "is_task" (bool) for task parameters.
                6. "digits" (int) for float-valued range parameters.
            parameter_constraints: List of string representation of parameter
                constraints, such as "x3 >= x4" or "-x3 + 2*x4 - 3.5*x5 >= 2". For
                the latter constraints, any number of arguments is accepted, and
                acceptable operators are "<=" and ">=". Note that parameter
                constraints may only be placed on range parameters, not choice
                parameters or fixed parameters.
        """
        self.experiment.search_space = self.make_search_space(
            parameters=parameters, parameter_constraints=parameter_constraints
        )
        self._save_experiment_to_db_if_possible(
            experiment=self.experiment,
        )

    @retry_on_exception(
        logger=logger,
        exception_types=(RuntimeError,),
        check_message_contains=["Cholesky", "cholesky"],
        suppress_all_errors=False,
        wrap_error_message_in=CHOLESKY_ERROR_ANNOTATION,
    )
    def get_next_trial(
        self,
        ttl_seconds: Optional[int] = None,
        force: bool = False,
        fixed_features: Optional[FixedFeatures] = None,
    ) -> Tuple[TParameterization, int]:
        """
        Generate trial with the next set of parameters to try in the iteration process.

        Note: Service API currently supports only 1-arm trials.

        Args:
            ttl_seconds: If specified, will consider the trial failed after this
                many seconds. Used to detect dead trials that were not marked
                failed properly.
            force: If set to True, this function will bypass the global stopping
                strategy's decision and generate a new trial anyway.
            fixed_features: A FixedFeatures object containing any
                features that should be fixed at specified values during
                generation.

        Returns:
            Tuple of trial parameterization, trial index
        """

        # Check if the global stopping strategy suggests to stop the optimization.
        # This is needed only if there is actually a stopping strategy specified,
        # and if this function is not forced to generate a new trial.
        if self.global_stopping_strategy and (not force):
            # The strategy itself will check if enough trials have already been
            # completed.
            (
                stop_optimization,
                global_stopping_message,
            ) = self.global_stopping_strategy.should_stop_optimization(
                experiment=self.experiment
            )
            if stop_optimization:
                raise OptimizationShouldStop(message=global_stopping_message)

        try:
            trial = self.experiment.new_trial(
                generator_run=self._gen_new_generator_run(
                    fixed_features=fixed_features
                ),
                ttl_seconds=ttl_seconds,
            )
        except MaxParallelismReachedException as e:
            if self._early_stopping_strategy is not None:
                e.message += (  # noqa: B306
                    " When stopping trials early, make sure to call `stop_trial_early` "
                    "on the stopped trial."
                )
            raise e
        logger.info(
            f"Generated new trial {trial.index} with parameters "
            f"{round_floats_for_logging(item=not_none(trial.arm).parameters)} "
            f"using model {not_none(trial.generator_run)._model_key}."
        )
        trial.mark_running(no_runner_required=True)
        self._save_or_update_trial_in_db_if_possible(
            experiment=self.experiment, trial=trial
        )
        # TODO[T79183560]: Ensure correct handling of generator run when using
        # foreign keys.
        self._update_generation_strategy_in_db_if_possible(
            generation_strategy=self.generation_strategy,
            new_generator_runs=[self.generation_strategy._generator_runs[-1]],
        )
        return not_none(trial.arm).parameters, trial.index

    def get_current_trial_generation_limit(self) -> Tuple[int, bool]:
        """How many trials this ``AxClient`` instance can currently produce via
        calls to ``get_next_trial``, before more trials are completed, and whether
        the optimization is complete.

        NOTE: If return value of this function is ``(0, False)``, no more trials
        can currently be procuded by this ``AxClient`` instance, but optimization
        is not completed; once more trials are completed with data, more new
        trials can be generated.

        Returns: a two-item tuple of:
              - the number of trials that can currently be produced, with -1
                meaning unlimited trials,
              - whether no more trials can be produced by this ``AxClient``
                instance at any point (e.g. if the search space is exhausted or
                generation strategy is completed.
        """
        # Ensure that experiment is set on the generation strategy.
        if self.generation_strategy._experiment is None:
            self.generation_strategy.experiment = self.experiment

        return self.generation_strategy.current_generator_run_limit()

    def get_next_trials(
        self,
        max_trials: int,
        ttl_seconds: Optional[int] = None,
        fixed_features: Optional[FixedFeatures] = None,
    ) -> Tuple[Dict[int, TParameterization], bool]:
        """Generate as many trials as currently possible.

        NOTE: Useful for running multiple trials in parallel: produces multiple trials,
        with their number limited by:
          - parallelism limit on current generation step,
          - number of trials in current generation step,
          - number of trials required to complete before moving to next generation step,
            if applicable,
          - and ``max_trials`` argument to this method.

        Args:
            max_trials: Limit on how many trials the call to this method should produce.
            ttl_seconds: If specified, will consider the trial failed after this
                many seconds. Used to detect dead trials that were not marked
                failed properly.
            fixed_features: A FixedFeatures object containing any
                features that should be fixed at specified values during
                generation.

        Returns: two-item tuple of:
              - mapping from trial indices to parameterizations in those trials,
              - boolean indicator of whether optimization is completed and no more
                trials can be generated going forward.
        """
        gen_limit, optimization_complete = self.get_current_trial_generation_limit()
        if optimization_complete:
            return {}, True

        # Trial generation limit of -1 indicates that unlimited trials can be
        # generated, so we only want to limit `max_trials` if `trial_generation_
        # limit` is non-negative.
        if gen_limit >= 0:
            max_trials = min(gen_limit, max_trials)

        trials_dict = {}
        for _ in range(max_trials):
            try:
                params, trial_index = self.get_next_trial(
                    ttl_seconds=ttl_seconds, fixed_features=fixed_features
                )
                trials_dict[trial_index] = params
            except OptimizationComplete as err:
                logger.info(
                    f"Encountered exception indicating optimization completion: {err}"
                )
                return trials_dict, True

        # Check whether optimization is complete now that we generated a batch
        # of trials.
        _, optimization_complete = self.get_current_trial_generation_limit()
        return trials_dict, optimization_complete

    def abandon_trial(self, trial_index: int, reason: Optional[str] = None) -> None:
        """Abandons a trial and adds optional metadata to it.

        Args:
            trial_index: Index of trial within the experiment.
        """
        trial = self.get_trial(trial_index)
        trial.mark_abandoned(reason=reason)

    def update_running_trial_with_intermediate_data(
        self,
        trial_index: int,
        raw_data: TEvaluationOutcome,
        metadata: Optional[Dict[str, Union[str, int]]] = None,
        sample_size: Optional[int] = None,
    ) -> None:
        """
        Updates the trial with given metric values without completing it. Also
        adds optional metadata to it. Useful for intermediate results like
        the metrics of a partially optimized machine learning model. In these
        cases it should be called instead of `complete_trial` until it is
        time to complete the trial.

        NOTE: This method will raise an Exception if it is called multiple times
        with the same ``raw_data``. These cases typically arise when ``raw_data``
        does not change over time. To avoid this, pass a timestep metric in
        ``raw_data``, for example:

        .. code-block:: python

            for ts in range(100):
                raw_data = [({"ts": ts}, {"my_objective": (1.0, 0.0)})]
                ax_client.update_running_trial_with_intermediate_data(
                    trial_index=0, raw_data=raw_data
                )

        NOTE: When ``raw_data`` does not specify SEM for a given metric, Ax
        will default to the assumption that the data is noisy (specifically,
        corrupted by additive zero-mean Gaussian noise) and that the
        level of noise should be inferred by the optimization model. To
        indicate that the data is noiseless, set SEM to 0.0, for example:

        .. code-block:: python

            ax_client.update_running_trial_with_intermediate_data(
                trial_index=0,
                raw_data={"my_objective": (objective_mean_value, 0.0)}
            )

        Args:
            trial_index: Index of trial within the experiment.
            raw_data: Evaluation data for the trial. Can be a mapping from
                metric name to a tuple of mean and SEM, just a tuple of mean and
                SEM if only one metric in optimization, or just the mean if SEM is
                unknown (then Ax will infer observation noise level).
                Can also be a list of (fidelities, mapping from
                metric name to a tuple of mean and SEM).
            metadata: Additional metadata to track about this run.
            sample_size: Number of samples collected for the underlying arm,
                optional.
        """
        if not isinstance(trial_index, int):
            raise ValueError(f"Trial index must be an int, got: {trial_index}.")
        if not self.experiment.default_data_type == DataType.MAP_DATA:
            raise ValueError(
                "`update_running_trial_with_intermediate_data` requires that "
                "this client's `experiment` be constructed with "
                "`support_intermediate_data=True` and have `default_data_type` of "
                "`DataType.MAP_DATA`."
            )
        data_update_repr = self._update_trial_with_raw_data(
            trial_index=trial_index,
            raw_data=raw_data,
            metadata=metadata,
            sample_size=sample_size,
            combine_with_last_data=True,
        )
        logger.info(f"Updated trial {trial_index} with data: " f"{data_update_repr}.")

    def complete_trial(
        self,
        trial_index: int,
        raw_data: TEvaluationOutcome,
        metadata: Optional[Dict[str, Union[str, int]]] = None,
        sample_size: Optional[int] = None,
    ) -> None:
        """
        Completes the trial with given metric values and adds optional metadata
        to it.

        NOTE: When ``raw_data`` does not specify SEM for a given metric, Ax
        will default to the assumption that the data is noisy (specifically,
        corrupted by additive zero-mean Gaussian noise) and that the
        level of noise should be inferred by the optimization model. To
        indicate that the data is noiseless, set SEM to 0.0, for example:

        .. code-block:: python

          ax_client.complete_trial(
              trial_index=0,
              raw_data={"my_objective": (objective_mean_value, 0.0)}
          )

        Args:
            trial_index: Index of trial within the experiment.
            raw_data: Evaluation data for the trial. Can be a mapping from
                metric name to a tuple of mean and SEM, just a tuple of mean and
                SEM if only one metric in optimization, or just the mean if SEM is
                unknown (then Ax will infer observation noise level).
                Can also be a list of (fidelities, mapping from
                metric name to a tuple of mean and SEM).
            metadata: Additional metadata to track about this run.
            sample_size: Number of samples collected for the underlying arm,
                optional.
        """
        # Validate that trial can be completed.
        trial = self.get_trial(trial_index)
        trial._validate_can_attach_data()
        if not isinstance(trial_index, int):
            raise ValueError(f"Trial index must be an int, got: {trial_index}.")
        data_update_repr = self._update_trial_with_raw_data(
            trial_index=trial_index,
            raw_data=raw_data,
            metadata=metadata,
            sample_size=sample_size,
            complete_trial=True,
            combine_with_last_data=True,
        )
        logger.info(f"Completed trial {trial_index} with data: " f"{data_update_repr}.")

    def update_trial_data(
        self,
        trial_index: int,
        raw_data: TEvaluationOutcome,
        metadata: Optional[Dict[str, Union[str, int]]] = None,
        sample_size: Optional[int] = None,
    ) -> None:
        """
        Attaches additional data or updates the existing data for a trial in a
        terminal state. For example, if trial was completed with data for only
        one of the required metrics, this can be used to attach data for the
        remaining metrics.

        NOTE: This does not change the trial status.

        Args:
            trial_index: Index of trial within the experiment.
            raw_data: Evaluation data for the trial. Can be a mapping from
                metric name to a tuple of mean and SEM, just a tuple of mean and
                SEM if only one metric in optimization, or just the mean if there
                is no SEM.  Can also be a list of (fidelities, mapping from
                metric name to a tuple of mean and SEM).
            metadata: Additional metadata to track about this run.
            sample_size: Number of samples collected for the underlying arm,
                optional.
        """
        if not isinstance(trial_index, int):
            raise ValueError(f"Trial index must be an int, got: {trial_index}.")
        trial = self.get_trial(trial_index)
        if not trial.status.is_terminal:
            raise ValueError(
                f"Trial {trial.index} is not in a terminal state. Use "
                "`ax_client.complete_trial` to complete the trial with new data "
                "or use `ax_client.update_running_trial_with_intermediate_data` "
                "to attach intermediate data to a running trial."
            )
        data_update_repr = self._update_trial_with_raw_data(
            trial_index=trial_index,
            raw_data=raw_data,
            metadata=metadata,
            sample_size=sample_size,
            combine_with_last_data=True,
        )
        logger.info(f"Added data: {data_update_repr} to trial {trial.index}.")

    def log_trial_failure(
        self, trial_index: int, metadata: Optional[Dict[str, str]] = None
    ) -> None:
        """Mark that the given trial has failed while running.

        Args:
            trial_index: Index of trial within the experiment.
            metadata: Additional metadata to track about this run.
        """
        trial = self.experiment.trials[trial_index]
        trial.mark_failed()
        logger.info(f"Registered failure of trial {trial_index}.")
        if metadata is not None:
            trial._run_metadata = metadata
        self._save_experiment_to_db_if_possible(
            experiment=self.experiment,
        )

    def attach_trial(
        self,
        parameters: TParameterization,
        ttl_seconds: Optional[int] = None,
        run_metadata: Optional[Dict[str, Any]] = None,
        arm_name: Optional[str] = None,
    ) -> Tuple[TParameterization, int]:
        """Attach a new trial with the given parameterization to the experiment.

        Args:
            parameters: Parameterization of the new trial.
            ttl_seconds: If specified, will consider the trial failed after this
                many seconds. Used to detect dead trials that were not marked
                failed properly.

        Returns:
            Tuple of parameterization and trial index from newly created trial.
        """

        output_parameters, trial_index = self.experiment.attach_trial(
            parameterizations=[parameters],
            arm_names=[arm_name] if arm_name else None,
            ttl_seconds=ttl_seconds,
            run_metadata=run_metadata,
        )
        self._save_or_update_trial_in_db_if_possible(
            experiment=self.experiment,
            trial=self.experiment.trials[trial_index],
        )

        return list(output_parameters.values())[0], trial_index

    def get_trial_parameters(self, trial_index: int) -> TParameterization:
        """Retrieve the parameterization of the trial by the given index."""
        return not_none(self.get_trial(trial_index).arm).parameters

    def get_trials_data_frame(self) -> pd.DataFrame:
        """Get a Pandas DataFrame representation of this experiment. The columns
        will include all the parameters in the search space and all the metrics
        on this experiment. The rows will each correspond to a trial (if using
        one-arm trials, which is the case in base ``AxClient``; will correspond
        to arms in trials in the batch-trial case).
        """
        return exp_to_df(exp=self.experiment)

    def get_max_parallelism(self) -> List[Tuple[int, int]]:
        """Retrieves maximum number of trials that can be scheduled in parallel
        at different stages of optimization.

        Some optimization algorithms profit significantly from sequential
        optimization (i.e. suggest a few points, get updated with data for them,
        repeat, see https://ax.dev/docs/bayesopt.html).
        Parallelism setting indicates how many trials should be running simulteneously
        (generated, but not yet completed with data).

        The output of this method is mapping of form
        {num_trials -> max_parallelism_setting}, where the max_parallelism_setting
        is used for num_trials trials. If max_parallelism_setting is -1, as
        many of the trials can be ran in parallel, as necessary. If num_trials
        in a tuple is -1, then the corresponding max_parallelism_setting
        should be used for all subsequent trials.

        For example, if the returned list is [(5, -1), (12, 6), (-1, 3)],
        the schedule could be: run 5 trials with any parallelism, run 6 trials in
        parallel twice, run 3 trials in parallel for as long as needed. Here,
        'running' a trial means obtaining a next trial from `AxClient` through
        get_next_trials and completing it with data when available.

        Returns:
            Mapping of form {num_trials -> max_parallelism_setting}.
        """
        parallelism_settings = []
        for step in self.generation_strategy._steps:
            parallelism_settings.append(
                (step.num_trials, step.max_parallelism or step.num_trials)
            )
        return parallelism_settings

    def get_optimization_trace(
        self, objective_optimum: Optional[float] = None
    ) -> AxPlotConfig:
        """Retrieves the plot configuration for optimization trace, which shows
        the evolution of the objective mean over iterations.

        Args:
            objective_optimum: Optimal objective, if known, for display in the
                visualization.
        """
        if not self.experiment.trials:
            raise ValueError("Cannot generate plot as there are no trials.")

        objective = self.objective
        if isinstance(objective, MultiObjective):
            raise UnsupportedError(
                "`get_optimization_trace` is not supported "
                "for multi-objective experiments"
            )

        # Setting the objective values of infeasible points to be infinitely
        # bad prevents them from increasing or decreasing the
        # optimization trace.
        def _constrained_trial_objective_mean(trial: BaseTrial) -> float:
            if constraint_satisfaction(trial):
                return checked_cast(Trial, trial).objective_mean
            return float("inf") if self.objective.minimize else float("-inf")

        objective_name = self.objective_name
        best_objectives = np.array(
            [
                [
                    _constrained_trial_objective_mean(trial)
                    for trial in self.experiment.trials.values()
                    if trial.status.is_completed
                ]
            ]
        )
        hover_labels = [
            _format_dict(not_none(checked_cast(Trial, trial).arm).parameters)
            for trial in self.experiment.trials.values()
            if trial.status.is_completed
        ]
        return optimization_trace_single_method(
            y=(
                np.minimum.accumulate(best_objectives, axis=1)
                if objective.minimize
                else np.maximum.accumulate(best_objectives, axis=1)
            ),
            optimum=objective_optimum,
            title="Best objective found vs. # of iterations",
            ylabel=objective_name.capitalize(),
            hover_labels=hover_labels,
        )

    def get_contour_plot(
        self,
        param_x: Optional[str] = None,
        param_y: Optional[str] = None,
        metric_name: Optional[str] = None,
    ) -> AxPlotConfig:
        """Retrieves a plot configuration for a contour plot of the response
        surface. For response surfaces with more than two parameters,
        selected two parameters will appear on the axes, and remaining parameters
        will be affixed to the middle of their range. If contour params arguments
        are not provided, the first two parameters in the search space will be
        used. If contour metrics are not provided, objective will be used.

        Args:
            param_x: name of parameters to use on x-axis for
                the contour response surface plots.
            param_y: name of parameters to use on y-axis for
                the contour response surface plots.
            metric_name: Name of the metric, for which to plot the response
                surface.
        """
        if not self.experiment.trials:
            raise ValueError("Cannot generate plot as there are no trials.")
        if len(self.experiment.parameters) < 2:
            raise ValueError(
                "Cannot create a contour plot as experiment has less than 2 "
                "parameters, but a contour-related argument was provided."
            )
        if (param_x or param_y) and not (param_x and param_y):
            raise ValueError(
                "If `param_x` is provided, `param_y` is "
                "required as well, and vice-versa."
            )

        if not metric_name:
            if isinstance(self.objective, MultiObjective):
                raise UnsupportedError(
                    "`get_contour_plot` requires a `metric_name` "
                    "for multi-objective experiments"
                )

            metric_name = self.objective_name

        if not param_x or not param_y:
            parameter_names = list(self.experiment.parameters.keys())
            param_x = parameter_names[0]
            param_y = parameter_names[1]

        if param_x not in self.experiment.parameters:
            raise ValueError(
                f'Parameter "{param_x}" not found in the optimization search space.'
            )
        if param_y not in self.experiment.parameters:
            raise ValueError(
                f'Parameter "{param_y}" not found in the optimization search space.'
            )
        if metric_name not in self.experiment.metrics:
            raise ValueError(
                f'Metric "{metric_name}" is not associated with this optimization.'
            )
        if self.generation_strategy.model is not None:
            try:
                logger.info(
                    f"Retrieving contour plot with parameter '{param_x}' on X-axis "
                    f"and '{param_y}' on Y-axis, for metric '{metric_name}'. "
                    "Remaining parameters are affixed to the middle of their range."
                )
                return plot_contour(
                    model=not_none(self.generation_strategy.model),
                    param_x=param_x,
                    param_y=param_y,
                    metric_name=metric_name,
                )

            except NotImplementedError:
                # Some models don't implement '_predict', which is needed
                # for the contour plots.
                logger.info(
                    f"Model {self.generation_strategy.model} does not implement "
                    "`predict`, so it cannot be used to generate a response "
                    "surface plot."
                )
        raise UnsupportedPlotError(
            f'Could not obtain contour plot of "{metric_name}" for parameters '
            f'"{param_x}" and "{param_y}", as a model with predictive ability, '
            "such as a Gaussian Process, has not yet been trained in the course "
            "of this optimization."
        )

    def get_feature_importances(self, relative: bool = True) -> AxPlotConfig:
        """
        Get a bar chart showing feature_importances for a metric.

        A drop-down controls the metric for which the importances are displayed.

        Args:
            relative: Whether the values are displayed as percentiles or
                as raw importance metrics.
        """
        if not self.experiment.trials:
            raise ValueError("Cannot generate plot as there are no trials.")
        cur_model = self.generation_strategy.model
        if cur_model is not None:
            try:
                return plot_feature_importance_by_feature(cur_model, relative=relative)
            except NotImplementedError:
                logger.info(
                    f"Model {self.generation_strategy.model} does not implement "
                    "`feature_importances`, so it cannot be used to generate "
                    "this plot. Only certain models, specifically GPEI, implement "
                    "feature importances."
                )

        raise ValueError(
            "Could not obtain feature_importances for any metrics "
            " as a model that can produce feature importances, such as a "
            "Gaussian Process, has not yet been trained in the course "
            "of this optimization."
        )

    def load_experiment_from_database(
        self,
        experiment_name: str,
        choose_generation_strategy_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Load an existing experiment from database using the `DBSettings`
        passed to this `AxClient` on instantiation.

        Args:
            experiment_name: Name of the experiment.

        Returns:
            Experiment object.
        """
        experiment, generation_strategy = self._load_experiment_and_generation_strategy(
            experiment_name=experiment_name
        )
        self._experiment = not_none(
            experiment, f"Experiment by name '{experiment_name}' not found."
        )
        logger.info(f"Loaded {experiment}.")
        if generation_strategy is None:
            self._set_generation_strategy(
                choose_generation_strategy_kwargs=choose_generation_strategy_kwargs
            )
            self._save_generation_strategy_to_db_if_possible()
        else:
            self._generation_strategy = generation_strategy
            logger.info(
                f"Using generation strategy associated with the loaded experiment:"
                f" {generation_strategy}."
            )

    def get_model_predictions_for_parameterizations(
        self,
        parameterizations: List[TParameterization],
        metric_names: Optional[List[str]] = None,
    ) -> List[Dict[str, Tuple[float, float]]]:
        """Retrieve model-estimated means and covariances for all metrics
        for the provided parameterizations.

        Args:
            metric_names: Names of the metrics for which to predict.
                All metrics will be predicted if this argument is
                not specified.
            parameterizations: List of Parameterizations for which to predict.

        Returns:
            A list of predicted metric mean and SEM of form:
            List[Tuple[float, float]].
        """

        parameterization_dict = dict(enumerate(parameterizations))

        predictions_dict = self.get_model_predictions(
            metric_names=metric_names, parameterizations=parameterization_dict
        )

        predictions_array = [
            predictions_dict[i] for i, _ in enumerate(parameterizations)
        ]

        return predictions_array

    def get_model_predictions(
        self,
        metric_names: Optional[List[str]] = None,
        include_out_of_sample: Optional[bool] = True,
        parameterizations: Optional[Dict[int, TParameterization]] = None,
    ) -> Dict[int, Dict[str, Tuple[float, float]]]:
        """Retrieve model-estimated means and covariances for all metrics.

        NOTE: This method currently only supports one-arm trials.

        Args:
            metric_names: Names of the metrics, for which to retrieve predictions.
                All metrics on experiment will be retrieved if this argument was
                not specified.
            include_out_of_sample: Defaults to True. Return predictions for
                out-of-sample (i.e. not yet completed trials) data in
                addition to in-sample (i.e. completed trials) data.
            parameterizations: Optional mapping from an int label to
                Parameterizations. When provided, predictions are performed *only*
                on these data points, no predictions from trial data is performed,
                and include_out_of_sample parameters is ignored.

        Returns:
            A mapping from trial index to a mapping of metric names to tuples
            of predicted metric mean and SEM, of form:
            { trial_index -> { metric_name: ( mean, SEM ) } }.
        """

        # Ensure there are metrics specified
        if metric_names is None and self.experiment.metrics is None:
            raise ValueError(
                "No metrics to retrieve specified on the experiment or as "
                "argument to `get_model_predictions`."
            )

        # Model update is normally tied to the GenerationStrategy.gen() call,
        # which is called from get_next_trial(). In order to ensure that predictions
        # can be performed without the need to call get_next_trial(), we update the
        # model with all attached data.
        self.fit_model()

        # Shared info for subsequent calls
        metric_names_to_predict = (
            set(metric_names)
            if metric_names is not None
            else set(not_none(self.experiment.metrics).keys())
        )
        model = not_none(
            self.generation_strategy.model, "No model has been instantiated yet."
        )

        # Construct a dictionary that maps from a label to an
        # ObservationFeature to predict.
        # - If returning trial predictions, the label is the trial index.
        # - If predictions are for user-provided parameterization, the label
        #   is provided in the input (also an int).
        label_to_feature_dict = {}

        # Predict on user-provided data
        if parameterizations is not None:
            logger.info(
                '"parameterizations" have been provided, only these data '
                "points will be predicted. No trial data prediction will be "
                "returned."
            )
            for label in parameterizations.keys():
                label_to_feature_dict[label] = ObservationFeatures(
                    parameters=parameterizations[label]
                )
        # Predict on associated trials
        else:
            trials_dict = self.experiment.trials
            for trial_index, trial in trials_dict.items():
                # filter trials based on input params and trial statuses
                if include_out_of_sample or trial.status.is_completed:
                    arms = trial.arms
                    if len(arms) > 1:
                        raise ValueError("Currently only 1-arm trials are supported.")
                    label_to_feature_dict[trial_index] = ObservationFeatures.from_arm(
                        arms[0]
                    )

        return predict_by_features(
            model=model,
            label_to_feature_dict=label_to_feature_dict,
            metric_names=metric_names_to_predict,
        )

    def fit_model(self) -> None:
        """Fit a model using data collected from the trials so far.

        This method will attempt to fit the same model that would be used for generating
        the next trial. The resulting model may be different from the model that was
        used to generate the last trial, if the generation node is ready to transition.

        This method rarely needs to be called by the user, because model-fitting is
        usually handled indirectly through ``AxClient.get_next_trial()``. This method
        instantiates a new model if none is yet available, which may be the case if
        no trials have been generated using a model-based method.

        NOTE: If the current generation node is not model-based, no model may be fit.
        """
        if not self.experiment.trial_indices_by_status[TrialStatus.COMPLETED]:
            raise DataRequiredError(
                "At least one trial must be completed with data to fit a model."
            )
        # Check if we should transition before generating the next candidate.
        self.generation_strategy._maybe_transition_to_next_node()
        self.generation_strategy._fit_current_model(data=None)

    def verify_trial_parameterization(
        self, trial_index: int, parameterization: TParameterization
    ) -> bool:
        """Whether the given parameterization matches that of the arm in the trial
        specified in the trial index.
        """
        return not_none(self.get_trial(trial_index).arm).parameters == parameterization

    def should_stop_trials_early(
        self, trial_indices: Set[int]
    ) -> Dict[int, Optional[str]]:
        """Evaluate whether to early-stop running trials.

        Args:
            trial_indices: Indices of trials to consider for early stopping.

        Returns:
            A dictionary mapping trial indices that should be early stopped to
            (optional) messages with the associated reason.
        """
        if self._early_stopping_strategy is None:
            logger.warning(
                "No early_stopping_strategy was passed to AxClient. "
                "Defaulting to never stopping any trials early."
            )
        return early_stopping_utils.should_stop_trials_early(
            early_stopping_strategy=self._early_stopping_strategy,
            trial_indices=trial_indices,
            experiment=self.experiment,
        )

    def stop_trial_early(self, trial_index: int) -> None:
        trial = self.get_trial(trial_index)
        trial.mark_early_stopped()
        logger.info(f"Early stopped trial {trial_index}.")
        self._save_or_update_trial_in_db_if_possible(
            experiment=self.experiment, trial=trial
        )

    def estimate_early_stopping_savings(self, map_key: Optional[str] = None) -> float:
        """Estimate early stopping savings using progressions of the MapMetric present
        on the EarlyStoppingConfig as a proxy for resource usage.

        Args:
            map_key: The name of the map_key by which to estimate early stopping
                savings, usually steps. If none is specified use some arbitrary map_key
                in the experiment's MapData

        Returns:
            The estimated resource savings as a fraction of total resource usage (i.e.
            0.11 estimated savings indicates we would expect the experiment to have used
            11% more resources without early stopping present)
        """
        if self.experiment.default_data_constructor is not MapData:
            return 0

        strategy = self.early_stopping_strategy
        map_key = (
            map_key
            if map_key is not None
            else (
                assert_is_instance(
                    self.experiment.metrics[list(strategy.metric_names)[0]],
                    MapMetric,
                ).map_key_info.key
                if strategy is not None
                and strategy.metric_names is not None
                and len(list(strategy.metric_names)) > 0
                else None
            )
        )

        return estimate_early_stopping_savings(
            experiment=self.experiment,
            map_key=map_key,
        )

    # ------------------ JSON serialization & storage methods. -----------------

    def save_to_json_file(self, filepath: str = "ax_client_snapshot.json") -> None:
        """Save a JSON-serialized snapshot of this `AxClient`'s settings and state
        to a .json file by the given path.
        """
        with open(filepath, "w+") as file:
            file.write(json.dumps(self.to_json_snapshot()))
            logger.info(f"Saved JSON-serialized state of optimization to `{filepath}`.")

    @classmethod
    def load_from_json_file(
        cls: Type[AxClientSubclass],
        filepath: str = "ax_client_snapshot.json",
        # pyre-fixme[2]: Parameter must be annotated.
        **kwargs,
    ) -> AxClientSubclass:
        """Restore an `AxClient` and its state from a JSON-serialized snapshot,
        residing in a .json file by the given path.
        """
        with open(filepath, "r") as file:
            serialized = json.loads(file.read())
            return cls.from_json_snapshot(serialized=serialized, **kwargs)

    def to_json_snapshot(
        self,
        # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        encoder_registry: Optional[Dict[Type, Callable[[Any], Dict[str, Any]]]] = None,
        # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        class_encoder_registry: Optional[
            Dict[Type, Callable[[Any], Dict[str, Any]]]
        ] = None,
    ) -> Dict[str, Any]:
        """Serialize this `AxClient` to JSON to be able to interrupt and restart
        optimization and save it to file by the provided path.

        Returns:
            A JSON-safe dict representation of this `AxClient`.
        """
        if encoder_registry is None:
            encoder_registry = CORE_ENCODER_REGISTRY

        if class_encoder_registry is None:
            class_encoder_registry = CORE_CLASS_ENCODER_REGISTRY

        return {
            "_type": self.__class__.__name__,
            "experiment": object_to_json(
                self._experiment,
                encoder_registry=encoder_registry,
                class_encoder_registry=class_encoder_registry,
            ),
            "generation_strategy": object_to_json(
                self._generation_strategy,
                encoder_registry=encoder_registry,
                class_encoder_registry=class_encoder_registry,
            ),
            "_enforce_sequential_optimization": self._enforce_sequential_optimization,
        }

    @classmethod
    def from_json_snapshot(
        cls: Type[AxClientSubclass],
        serialized: Dict[str, Any],
        decoder_registry: Optional[TDecoderRegistry] = None,
        # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
        class_decoder_registry: Optional[
            Dict[str, Callable[[Dict[str, Any]], Any]]
        ] = None,
        # pyre-fixme[2]: Parameter must be annotated.
        **kwargs,
    ) -> AxClientSubclass:
        """Recreate an `AxClient` from a JSON snapshot."""
        if decoder_registry is None:
            decoder_registry = CORE_DECODER_REGISTRY

        if class_decoder_registry is None:
            class_decoder_registry = CORE_CLASS_DECODER_REGISTRY

        experiment = object_from_json(
            serialized.pop("experiment"),
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        )
        serialized_generation_strategy = serialized.pop("generation_strategy")
        ax_client = cls(
            generation_strategy=(
                generation_strategy_from_json(
                    generation_strategy_json=serialized_generation_strategy,
                    experiment=experiment,
                    decoder_registry=decoder_registry,
                    class_decoder_registry=class_decoder_registry,
                )
                if serialized_generation_strategy is not None
                else None
            ),
            enforce_sequential_optimization=serialized.pop(
                "_enforce_sequential_optimization"
            ),
            **kwargs,
        )
        ax_client._experiment = experiment
        return ax_client

    # ---------------------- Private helper methods. ---------------------

    @property
    def experiment(self) -> Experiment:
        """Returns the experiment set on this Ax client."""
        return not_none(
            self._experiment,
            (
                "Experiment not set on Ax client. Must first "
                "call load_experiment or create_experiment to use handler functions."
            ),
        )

    def get_trial(self, trial_index: int) -> Trial:
        """Return a trial on experiment cast as Trial"""
        return checked_cast(Trial, self.experiment.trials[trial_index])

    @property
    def generation_strategy(self) -> GenerationStrategy:
        """Returns the generation strategy, set on this experiment."""
        return not_none(
            self._generation_strategy,
            "No generation strategy has been set on this optimization yet.",
        )

    @property
    def objective(self) -> Objective:
        return not_none(self.experiment.optimization_config).objective

    @property
    def objective_name(self) -> str:
        """Returns the name of the objective in this optimization."""
        objective = self.objective
        if isinstance(objective, MultiObjective):
            raise UnsupportedError(
                "Multi-objective experiments contain multiple objectives"
            )
        return objective.metric.name

    @property
    def objective_names(self) -> List[str]:
        """Returns the name of the objective in this optimization."""
        objective = self.objective
        return [m.name for m in objective.metrics]

    @property
    def metric_definitions(self) -> Dict[str, Dict[str, Any]]:
        """Returns metric definitions for all experiment metrics that can
        be passed into functions requiring metric_definitions
        """
        return {
            m.serialize_init_args(m)["name"]: {
                "metric_class": m.__class__,
                **{k: v for k, v in m.serialize_init_args(m).items() if k != "name"},
            }
            for m in self.experiment.metrics.values()
        }

    @property
    def metric_names(self) -> Set[str]:
        """Returns the names of all metrics on the attached experiment."""
        return set(self.experiment.metrics)

    @property
    def early_stopping_strategy(self) -> Optional[BaseEarlyStoppingStrategy]:
        """The early stopping strategy used on the experiment."""
        return self._early_stopping_strategy

    @early_stopping_strategy.setter
    def early_stopping_strategy(self, ess: BaseEarlyStoppingStrategy) -> None:
        """Update the early stopping strategy."""
        self._early_stopping_strategy = ess

    @property
    def global_stopping_strategy(self) -> Optional[BaseGlobalStoppingStrategy]:
        """The global stopping strategy used on the experiment."""
        return self._global_stopping_strategy

    @global_stopping_strategy.setter
    def global_stopping_strategy(self, gss: BaseGlobalStoppingStrategy) -> None:
        """Update the global stopping strategy."""
        self._global_stopping_strategy = gss

    @copy_doc(BestPointMixin.get_best_trial)
    def get_best_trial(
        self,
        optimization_config: Optional[OptimizationConfig] = None,
        trial_indices: Optional[Iterable[int]] = None,
        use_model_predictions: bool = True,
    ) -> Optional[Tuple[int, TParameterization, Optional[TModelPredictArm]]]:
        return self._get_best_trial(
            experiment=self.experiment,
            generation_strategy=self.generation_strategy,
            trial_indices=trial_indices,
            use_model_predictions=use_model_predictions,
        )

    @copy_doc(BestPointMixin.get_pareto_optimal_parameters)
    def get_pareto_optimal_parameters(
        self,
        optimization_config: Optional[OptimizationConfig] = None,
        trial_indices: Optional[Iterable[int]] = None,
        use_model_predictions: bool = True,
    ) -> Dict[int, Tuple[TParameterization, TModelPredictArm]]:
        return self._get_pareto_optimal_parameters(
            experiment=self.experiment,
            generation_strategy=self.generation_strategy,
            trial_indices=trial_indices,
            use_model_predictions=use_model_predictions,
        )

    @copy_doc(BestPointMixin.get_hypervolume)
    def get_hypervolume(
        self,
        optimization_config: Optional[MultiObjectiveOptimizationConfig] = None,
        trial_indices: Optional[Iterable[int]] = None,
        use_model_predictions: bool = True,
    ) -> float:
        return BestPointMixin._get_hypervolume(
            experiment=self.experiment,
            generation_strategy=self.generation_strategy,
            optimization_config=optimization_config,
            trial_indices=trial_indices,
            use_model_predictions=use_model_predictions,
        )

    @copy_doc(BestPointMixin.get_trace)
    def get_trace(
        self,
        optimization_config: Optional[MultiObjectiveOptimizationConfig] = None,
    ) -> List[float]:
        return BestPointMixin._get_trace(
            experiment=self.experiment,
            optimization_config=optimization_config,
        )

    @copy_doc(BestPointMixin.get_trace_by_progression)
    def get_trace_by_progression(
        self,
        optimization_config: Optional[OptimizationConfig] = None,
        bins: Optional[List[float]] = None,
        final_progression_only: bool = False,
    ) -> Tuple[List[float], List[float]]:
        return BestPointMixin._get_trace_by_progression(
            experiment=self.experiment,
            optimization_config=optimization_config,
            bins=bins,
            final_progression_only=final_progression_only,
        )

    def _update_trial_with_raw_data(
        self,
        trial_index: int,
        raw_data: TEvaluationOutcome,
        metadata: Optional[Dict[str, Union[str, int]]] = None,
        sample_size: Optional[int] = None,
        complete_trial: bool = False,
        combine_with_last_data: bool = False,
    ) -> str:
        """Helper method attaches data to a trial, returns a str of update."""
        # Format the data to save.
        trial = self.get_trial(trial_index)
        update_info = trial.update_trial_data(
            raw_data=raw_data,
            metadata=metadata,
            sample_size=sample_size,
            combine_with_last_data=combine_with_last_data,
        )

        if complete_trial:
            if not self._validate_all_required_metrics_present(
                raw_data=raw_data, trial_index=trial_index
            ):
                logger.warning(
                    "Marking the trial as failed because it is missing one"
                    "or more required metrics."
                )
                trial.mark_failed()
            else:
                trial.mark_completed()

        self._save_or_update_trial_in_db_if_possible(
            experiment=self.experiment, trial=trial
        )

        return update_info

    def _set_experiment(
        self,
        experiment: Experiment,
        overwrite_existing_experiment: bool = False,
    ) -> None:
        """Sets the ``_experiment`` attribute on this `AxClient`` instance and saves the
        experiment if this instance uses SQL storage.

        NOTE: This setter **should not be used outside of this file in production**.
        It can be leveraged in development, but all checked-in code that uses the
        Service API should leverage ``AxClient.create_experiment`` instead and extend it
        as needed. If using ``create_experiment`` is impossible and this setter is
        required, please raise your use case to the AE team or on our Github.
        """
        name = experiment._name

        if self.db_settings_set and not name:
            raise ValueError(
                "Must give the experiment a name if `db_settings` is not None."
            )
        if self.db_settings_set:
            experiment_id, _ = self._get_experiment_and_generation_strategy_db_id(
                experiment_name=not_none(name)
            )
            if experiment_id:
                raise ValueError(
                    f"Experiment {name} already exists in the database. "
                    "To protect experiments that are running in production, "
                    "overwriting stored experiments is not allowed. To "
                    "start a new experiment and store it, change the "
                    "experiment's name."
                )
        if self._experiment is not None:
            if overwrite_existing_experiment:
                exp_name = self.experiment._name or "untitled"
                new_exp_name = name or "untitled"
                logger.info(
                    f"Overwriting existing experiment ({exp_name}) on this client "
                    f"with new experiment ({new_exp_name}) and restarting the "
                    "generation strategy."
                )
                self._generation_strategy = None
            else:
                raise ValueError(
                    "Experiment already created for this client instance. "
                    "Set the `overwrite_existing_experiment` to `True` to overwrite "
                    "with new experiment."
                )

        self._experiment = experiment

        try:
            self._save_experiment_to_db_if_possible(
                experiment=self.experiment,
            )
        except Exception:
            # Unset the experiment on this `AxClient` instance if encountered and
            # raising an error from saving the experiment, to avoid a case where
            # overall `create_experiment` call fails with a storage error, but
            # `self._experiment` is still set and user has to specify the
            # `overwrite_existing_experiment` kwarg to re-attempt exp. creation.
            self._experiment = None
            raise

    def _set_runner(self, experiment: Experiment) -> None:
        """Overridable method to sets a runner on the experiment."""
        experiment.runner = None

    def _set_generation_strategy(
        self, choose_generation_strategy_kwargs: Optional[Dict[str, Any]] = None
    ) -> None:
        """Selects the generation strategy and applies specified dispatch kwargs,
        if any.
        """
        choose_generation_strategy_kwargs = choose_generation_strategy_kwargs or {}
        if (
            "use_batch_trials" in choose_generation_strategy_kwargs
            and type(self) is AxClient
        ):
            logger.warning(
                "Selecting a GenerationStrategy when using BatchTrials is in beta. "
                "Double check the recommended strategy matches your expectations."
            )
        random_seed = choose_generation_strategy_kwargs.pop(
            "random_seed", self._random_seed
        )
        enforce_sequential_optimization = choose_generation_strategy_kwargs.pop(
            "enforce_sequential_optimization", self._enforce_sequential_optimization
        )
        if self._generation_strategy is None:
            self._generation_strategy = choose_generation_strategy(
                search_space=self.experiment.search_space,
                optimization_config=self.experiment.optimization_config,
                enforce_sequential_optimization=enforce_sequential_optimization,
                random_seed=random_seed,
                torch_device=self._torch_device,
                experiment=self.experiment,
                **choose_generation_strategy_kwargs,
            )
        elif self._experiment:
            self._generation_strategy.experiment = self.experiment

    def _save_generation_strategy_to_db_if_possible(
        self,
        generation_strategy: Optional[GenerationStrategyInterface] = None,
    ) -> bool:
        return super()._save_generation_strategy_to_db_if_possible(
            generation_strategy=generation_strategy or self.generation_strategy,
        )

    def _gen_new_generator_run(
        self, n: int = 1, fixed_features: Optional[FixedFeatures] = None
    ) -> GeneratorRun:
        """Generate new generator run for this experiment.

        Args:
            n: Number of arms to generate.
            fixed_features: A FixedFeatures object containing any
                features that should be fixed at specified values during
                generation.
        """
        # If random seed is not set for this optimization, context manager does
        # nothing; otherwise, it sets the random seed for torch, but only for the
        # scope of this call. This is important because torch seed is set globally,
        # so if we just set the seed without the context manager, it can have
        # serious negative impact on the performance of the models that employ
        # stochasticity.

        fixed_feats = (
            InstantiationBase.make_fixed_observation_features(
                fixed_features=fixed_features
            )
            if fixed_features
            else None
        )
        with with_rng_seed(seed=self._random_seed):
            return not_none(self.generation_strategy).gen(
                experiment=self.experiment,
                n=n,
                pending_observations=self._get_pending_observation_features(
                    experiment=self.experiment
                ),
                fixed_features=fixed_feats,
            )

    def _find_last_trial_with_parameterization(
        self, parameterization: TParameterization
    ) -> int:
        """Given a parameterization, find the last trial in the experiment that
        contains an arm with that parameterization.
        """
        for trial_idx in sorted(self.experiment.trials.keys(), reverse=True):
            if not_none(self.get_trial(trial_idx).arm).parameters == parameterization:
                return trial_idx
        raise ValueError(
            f"No trial on experiment matches parameterization {parameterization}."
        )

    def _validate_all_required_metrics_present(
        self, raw_data: TEvaluationOutcome, trial_index: int
    ) -> bool:
        """Check if all required metrics are present in the given raw data."""
        opt_config = self.experiment.optimization_config
        if opt_config is None:
            return True

        _, data = data_and_evaluations_from_raw_data(
            raw_data={"data": raw_data},
            sample_sizes={},
            trial_index=trial_index,
            data_type=self.experiment.default_data_type,
            metric_names=opt_config.objective.metric_names,
        )
        required_metrics = set(opt_config.metrics.keys())
        provided_metrics = data.metric_names
        missing_metrics = required_metrics - provided_metrics
        return not missing_metrics

    @classmethod
    def _get_pending_observation_features(
        cls,
        experiment: Experiment,
    ) -> Optional[Dict[str, List[ObservationFeatures]]]:
        """Extract pending points for the given experiment.

        NOTE: With one-arm `Trial`-s, we use a more performant
        ``get_pending_observation_features_based_on_trial_status`` utility instead
        of ``get_pending_observation_features``, since we can determine whether a point
        is pending based on the status of the corresponding trial.
        """
        return get_pending_observation_features_based_on_trial_status(
            experiment=experiment
        )

    # ------------------------------ Validators. -------------------------------

    def _validate_early_stopping_strategy(
        self, support_intermediate_data: bool
    ) -> None:
        if self._early_stopping_strategy is not None and not support_intermediate_data:
            raise ValueError(
                "Early stopping is only supported for experiments which allow "
                " reporting intermediate trial data by setting passing "
                "`support_intermediate_data=True`."
            )

    def __repr__(self) -> str:
        """String representation of this client."""
        return f"{self.__class__.__name__}(experiment={self._experiment})"

    # -------- Backward-compatibility with old save / load method names. -------

    @staticmethod
    def get_recommended_max_parallelism() -> None:
        raise NotImplementedError(
            "Use `get_max_parallelism` instead; parallelism levels are now "
            "enforced in generation strategy, so max parallelism is no longer "
            "just recommended."
        )

    @staticmethod
    def load_experiment(experiment_name: str) -> None:
        raise NotImplementedError(
            "Use `load_experiment_from_database` to load from SQL database or "
            "`load_from_json_file` to load optimization state from .json file."
        )

    @staticmethod
    def load(filepath: Optional[str] = None) -> None:
        raise NotImplementedError(
            "Use `load_experiment_from_database` to load from SQL database or "
            "`load_from_json_file` to load optimization state from .json file."
        )

    @staticmethod
    def save(filepath: Optional[str] = None) -> None:
        raise NotImplementedError(
            "Use `save_to_json_file` to save optimization state to .json file."
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy

from ax.core.observation import ObservationFeatures
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.search_space import RobustSearchSpace, SearchSpace
from ax.modelbridge.transforms.remove_fixed import RemoveFixed
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_robust_search_space


class RemoveFixedTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "a", lower=1, upper=3, parameter_type=ParameterType.FLOAT
                ),
                ChoiceParameter(
                    "b", parameter_type=ParameterType.STRING, values=["a", "b", "c"]
                ),
                FixedParameter("c", parameter_type=ParameterType.STRING, value="a"),
            ]
        )
        self.t = RemoveFixed(
            search_space=self.search_space,
            observations=[],
        )

    def test_Init(self) -> None:
        self.assertEqual(list(self.t.fixed_parameters.keys()), ["c"])

    def test_TransformObservationFeatures(self) -> None:
        observation_features = [
            ObservationFeatures(parameters={"a": 2.2, "b": "b", "c": "a"})
        ]
        obs_ft2 = deepcopy(observation_features)
        obs_ft2 = self.t.transform_observation_features(obs_ft2)
        self.assertEqual(
            obs_ft2, [ObservationFeatures(parameters={"a": 2.2, "b": "b"})]
        )
        obs_ft2 = self.t.untransform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, observation_features)

        observation_features = [
            ObservationFeatures(parameters={"a": 2.2, "b": "b", "c": "a"})
        ]
        observation_features_invalid = [
            ObservationFeatures(parameters={"a": 2.2, "b": "b", "c": "b"})
        ]
        # Fixed parameter out of design!
        with self.assertRaises(ValueError):
            self.t.transform_observation_features(observation_features_invalid)

    def test_TransformSearchSpace(self) -> None:
        ss2 = self.search_space.clone()
        ss2 = self.t.transform_search_space(ss2)
        self.assertEqual(ss2.parameters.get("c"), None)

    def test_w_parameter_distributions(self) -> None:
        rss = get_robust_search_space()
        rss.add_parameter(
            FixedParameter("d", parameter_type=ParameterType.STRING, value="a"),
        )
        # Transform a non-distributional parameter.
        t = RemoveFixed(
            search_space=rss,
            observations=[],
        )
        rss = t.transform_search_space(rss)
        # Make sure that the return value is still a RobustSearchSpace.
        self.assertIsInstance(rss, RobustSearchSpace)
        self.assertEqual(len(rss.parameters.keys()), 4)
        # pyre-fixme[16]: `SearchSpace` has no attribute `parameter_distributions`.
        self.assertEqual(len(rss.parameter_distributions), 2)
        self.assertNotIn("d", rss.parameters)
        # Test with environmental variables.
        all_params = list(rss.parameters.values())
        rss = RobustSearchSpace(
            parameters=all_params[2:],
            parameter_distributions=rss.parameter_distributions,
            # pyre-fixme[16]: `SearchSpace` has no attribute `num_samples`.
            num_samples=rss.num_samples,
            environmental_variables=all_params[:2],
        )
        rss.add_parameter(
            FixedParameter("d", parameter_type=ParameterType.STRING, value="a"),
        )
        t = RemoveFixed(
            search_space=rss,
            observations=[],
        )
        rss = t.transform_search_space(rss)
        self.assertIsInstance(rss, RobustSearchSpace)
        self.assertEqual(len(rss.parameters.keys()), 4)
        self.assertEqual(len(rss.parameter_distributions), 2)
        # pyre-fixme[16]: `SearchSpace` has no attribute `_environmental_variables`.
        self.assertEqual(len(rss._environmental_variables), 2)
        self.assertNotIn("d", rss.parameters)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
import plotly.graph_objects as go
from ax.modelbridge.registry import Models
from ax.plot.base import AxPlotConfig
from ax.plot.trace import (
    optimization_trace_single_method,
    optimization_trace_single_method_plotly,
    plot_objective_value_vs_trial_index,
)
from ax.service.utils.report_utils import exp_to_df
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.mock import fast_botorch_optimize


class TracesTest(TestCase):
    @fast_botorch_optimize
    def setUp(self) -> None:
        super().setUp()
        self.exp = get_branin_experiment(with_batch=True)
        self.exp.trials[0].run()
        self.model = Models.BOTORCH_MODULAR(
            # Model bridge kwargs
            experiment=self.exp,
            data=self.exp.fetch_data(),
        )

    def test_Traces(self) -> None:
        # Assert that each type of plot can be constructed successfully
        plot = optimization_trace_single_method_plotly(
            np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
            list(self.model.metric_names)[0],
            optimization_direction="minimize",
            autoset_axis_limits=False,
        )
        self.assertIsInstance(plot, go.Figure)
        plot = optimization_trace_single_method(
            np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
            list(self.model.metric_names)[0],
            optimization_direction="minimize",
            autoset_axis_limits=False,
        )
        self.assertIsInstance(plot, AxPlotConfig)

    def test_TracesAutoAxes(self) -> None:
        for optimization_direction in ["minimize", "maximize", "passthrough"]:
            plot = optimization_trace_single_method_plotly(
                np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
                list(self.model.metric_names)[0],
                optimization_direction=optimization_direction,
                autoset_axis_limits=True,
            )
            self.assertIsNone(plot.layout.xaxis.range)  # pyre-ignore
            if optimization_direction == "minimize":
                self.assertAlmostEqual(plot.layout.yaxis.range[0], 0.525)
                self.assertAlmostEqual(plot.layout.yaxis.range[1], 6.225)
            elif optimization_direction == "maximize":
                self.assertAlmostEqual(plot.layout.yaxis.range[0], 0.775)
                self.assertAlmostEqual(plot.layout.yaxis.range[1], 6.475)
            else:
                self.assertIsNone(plot.layout.yaxis.range)

    @fast_botorch_optimize
    def test_plot_objective_value_vs_trial_index(self) -> None:
        # Generate some trials with different model types, including batch trial.
        exp = get_branin_experiment(with_batch=True)
        exp.trials[0].mark_completed(unsafe=True)
        sobol = Models.SOBOL(search_space=exp.search_space)
        for _ in range(2):
            t = exp.new_trial(sobol.gen(1)).run()
            t.mark_completed()
        model = Models.GPEI(
            experiment=exp,
            data=exp.fetch_data(),
        )
        for _ in range(2):
            t = exp.new_trial(model.gen(1)).run()
            t.mark_completed()
        exp.fetch_data()

        # Create exp_df and add feasibility column.
        exp_df = exp_to_df(exp)
        exp_df["is_feasible"] = True
        exp_df.loc[0, "is_feasible"] = False

        # Assert that plot can be constructed successfully.
        plot = plot_objective_value_vs_trial_index(
            exp_df=exp_df,
            metric_colname=list(self.model.metric_names)[0],
            minimize=True,
            hover_data_colnames=["trial_index"],
        )
        self.assertIsInstance(plot, go.Figure)

        # Assert that plot can be constructed successfully
        # without feasibility and generation method columns.
        del exp_df["is_feasible"]
        del exp_df["generation_method"]
        plot = plot_objective_value_vs_trial_index(
            exp_df=exp_df,
            metric_colname=list(self.model.metric_names)[0],
            minimize=True,
            hover_data_colnames=["trial_index"],
        )
        self.assertIsInstance(plot, go.Figure)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import enum
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Hashable,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np


TNumeric = Union[float, int]
TParamCounter = DefaultDict[int, int]
TParamValue = Union[None, str, bool, float, int]
TParameterization = Dict[str, TParamValue]
TParamValueList = List[TParamValue]  # a parameterization without the keys
TContextStratum = Optional[Dict[str, Union[str, float, int]]]

TBounds = Optional[Tuple[np.ndarray, np.ndarray]]
TModelMean = Dict[str, List[float]]
TModelCov = Dict[str, Dict[str, List[float]]]
TModelPredict = Tuple[TModelMean, TModelCov]
# Model predictions for a single arm:
# ( { metric -> mean }, { metric -> { other_metric -> covariance } } ).
TModelPredictArm = Tuple[Dict[str, float], Optional[Dict[str, Dict[str, float]]]]

FloatLike = Union[int, float, np.floating, np.integer]
SingleMetricDataTuple = Tuple[FloatLike, Optional[FloatLike]]
SingleMetricData = Union[FloatLike, Tuple[FloatLike, Optional[FloatLike]]]
# 1-arm `Trial` evaluation data: {metric_name -> (mean, standard error)}}.
TTrialEvaluation = Dict[str, SingleMetricData]

# 1-arm evaluation data with trace fidelities
TFidelityTrialEvaluation = List[Tuple[TParameterization, TTrialEvaluation]]

# 1-arm evaluation data with arbitrary partial results
TMapDict = Dict[str, Hashable]
TMapTrialEvaluation = List[Tuple[TMapDict, TTrialEvaluation]]

# Format for trasmitting evaluation data to Ax is either:
# 1) {metric_name -> (mean, standard error)} (TTrialEvaluation)
# 2) (mean, standard error) and we assume metric name == objective name
# 3) only the mean, and we assume metric name == objective name and standard error == 0
# 4) [({fidelity_param -> value}, {metric_name} -> (mean, standard error))]

TEvaluationOutcome = Union[
    TTrialEvaluation,
    SingleMetricData,
    TFidelityTrialEvaluation,
    TMapTrialEvaluation,
]
TEvaluationFunction = Union[
    Callable[[TParameterization], TEvaluationOutcome],
    Callable[[TParameterization, Optional[float]], TEvaluationOutcome],
]

TBucket = List[Dict[str, List[str]]]

TGenMetadata = Dict[str, Any]

# Model's metadata about a given candidate (or X).
TCandidateMetadata = Optional[Dict[str, Any]]


class ComparisonOp(enum.Enum):
    """Class for enumerating comparison operations."""

    GEQ: int = 0
    LEQ: int = 1


def merge_model_predict(
    predict: TModelPredict, predict_append: TModelPredict
) -> TModelPredict:
    """Append model predictions to an existing set of model predictions.

    TModelPredict is of the form:
        {metric_name: [mean1, mean2, ...],
        {metric_name: {metric_name: [var1, var2, ...]}})

    This will append the predictions

    Args:
        predict: Initial set of predictions.
        other_predict: Predictions to be appended.

    Returns:
        TModelPredict with the new predictions appended.
    """
    mu, cov = predict
    mu_append, cov_append = predict_append
    if len(mu) != len(mu_append) or len(cov) != len(cov_append):
        raise ValueError("Both sets of model predictions must have the same metrics")

    # Iterate down to the list level and simply add.
    for metric_name, metric_values in mu.items():
        mu[metric_name] = metric_values + mu_append[metric_name]

    for metric_name, co_cov in cov.items():
        for co_metric_name, cov_values in co_cov.items():
            cov[metric_name][co_metric_name] = (
                cov_values + cov_append[metric_name][co_metric_name]
            )
    return mu, cov


def validate_floatlike(floatlike: FloatLike) -> None:
    if not (
        isinstance(floatlike, float)
        or isinstance(floatlike, int)
        or isinstance(floatlike, np.floating)
        or isinstance(floatlike, np.integer)
    ):
        raise TypeError(f"Expected FloatLike, found {floatlike}")


def validate_single_metric_data(data: SingleMetricData) -> None:
    if isinstance(data, tuple):
        if len(data) != 2:
            raise TypeError(
                f"Tuple-valued SingleMetricData must have len == 2, found {data}"
            )

        mean, sem = data
        validate_floatlike(floatlike=mean)

        if sem is not None:
            validate_floatlike(floatlike=sem)

    else:
        validate_floatlike(floatlike=data)


def validate_trial_evaluation(evaluation: TTrialEvaluation) -> None:
    for key, value in evaluation.items():
        if not isinstance(key, str):
            raise TypeError(f"Keys must be strings in TTrialEvaluation, found {key}.")

        validate_single_metric_data(data=value)


def validate_param_value(param_value: TParamValue) -> None:
    if not (
        isinstance(param_value, str)
        or isinstance(param_value, bool)
        or isinstance(param_value, float)
        or isinstance(param_value, int)
        or param_value is None
    ):
        raise TypeError(f"Expected None, bool, float, int, or str, found {param_value}")


def validate_parameterization(parameterization: TParameterization) -> None:
    for key, value in parameterization.items():
        if not isinstance(key, str):
            raise TypeError(f"Keys must be strings in TParameterization, found {key}.")

        validate_param_value(param_value=value)


def validate_map_dict(map_dict: TMapDict) -> None:
    for key, value in map_dict.items():
        if not isinstance(key, str):
            raise TypeError(f"Keys must be strings in TMapDict, found {key}.")

        if not isinstance(value, Hashable):
            raise TypeError(f"Values must be Hashable in TMapDict, found {value}.")


def validate_fidelity_trial_evaluation(evaluation: TFidelityTrialEvaluation) -> None:
    for parameterization, trial_evaluation in evaluation:
        validate_parameterization(parameterization=parameterization)
        validate_trial_evaluation(evaluation=trial_evaluation)


def validate_map_trial_evaluation(evaluation: TMapTrialEvaluation) -> None:
    for map_dict, trial_evaluation in evaluation:
        validate_map_dict(map_dict=map_dict)
        validate_trial_evaluation(evaluation=trial_evaluation)


def validate_evaluation_outcome(outcome: TEvaluationOutcome) -> None:
    """Runtime validate that the supplied outcome has correct structure."""

    if isinstance(outcome, dict):
        # Check if outcome is TTrialEvaluation
        validate_trial_evaluation(evaluation=outcome)

    elif isinstance(outcome, list):
        # Check if outcome is TFidelityTrialEvaluation or TMapTrialEvaluation
        try:
            validate_fidelity_trial_evaluation(evaluation=outcome)  # pyre-ignore[6]
        except Exception:
            try:
                validate_map_trial_evaluation(evaluation=outcome)  # pyre-ignore[6]
            except Exception:
                raise TypeError(
                    "Expected either TFidelityTrialEvaluation or TMapTrialEvaluation, "
                    f"found {outcome}"
                )

    else:
        # Check if outcome is SingleMetricData
        validate_single_metric_data(data=outcome)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.core.metric import Metric, MetricFetchE
from ax.utils.common.result import Err
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_metric,
    get_data,
    get_experiment,
    get_factorial_metric,
)


class TestMetric(Metric):
    __test__ = False

    pass


METRIC_STRING = "Metric('m1')"


class MetricTest(TestCase):
    def setUp(self) -> None:
        pass

    def test_init(self) -> None:
        metric = Metric(name="m1", lower_is_better=False)
        self.assertEqual(str(metric), METRIC_STRING)

    def test_eq(self) -> None:
        metric1 = Metric(name="m1", lower_is_better=False)
        metric2 = Metric(name="m1", lower_is_better=False)
        self.assertEqual(metric1, metric2)

        metric3 = Metric(name="m1", lower_is_better=True)
        self.assertNotEqual(metric1, metric3)

    def test_clone(self) -> None:
        metric1 = Metric(name="m1", lower_is_better=False)
        self.assertEqual(metric1, metric1.clone())

        metric2 = get_branin_metric(name="branin")
        self.assertEqual(metric2, metric2.clone())

        metric3 = get_factorial_metric(name="factorial")
        self.assertEqual(metric3, metric3.clone())

    def test_sortable(self) -> None:
        metric1 = Metric(name="m1", lower_is_better=False)
        metric2 = Metric(name="m2", lower_is_better=False)
        self.assertTrue(metric1 < metric2)

    def test_wrap_unwrap(self) -> None:
        data = get_data()

        trial_multi = Metric._unwrap_trial_data_multi(
            results=Metric._wrap_trial_data_multi(data=data)
        )
        self.assertEqual(trial_multi, data)

        experiment = Metric._unwrap_experiment_data(
            results=Metric._wrap_experiment_data(data=data)
        )
        self.assertEqual(experiment, data)

        experiment_multi = Metric._unwrap_experiment_data_multi(
            results=Metric._wrap_experiment_data_multi(data=data)
        )
        self.assertEqual(experiment_multi, data)

    def test_wrap_err(self) -> None:
        err = Err(MetricFetchE(message="failed!", exception=Exception("panic!")))

        with self.assertRaisesRegex(Exception, "panic"):
            Metric._unwrap_experiment_data_multi(results={0: {"foo": err}})

        with self.assertRaisesRegex(Exception, "panic"):
            Metric._unwrap_experiment_data(results={0: err})

        with self.assertRaisesRegex(Exception, "panic"):
            Metric._unwrap_trial_data_multi(results={"foo": err})

    def test_MetricFetchE(self) -> None:
        def foo() -> bool:
            raise ValueError("bad value")

        def bar() -> bool:
            return foo()

        exception = None
        try:
            bar()
        except Exception as e:
            exception = e

        metric_fetch_e = MetricFetchE(message="foo", exception=exception)

        self.assertEqual(metric_fetch_e.message, "foo")

        self.assertIn("in foo", metric_fetch_e.__repr__())
        self.assertIn("in bar", metric_fetch_e.__repr__())
        self.assertIn("ValueError: bad value", metric_fetch_e.__repr__())

    def test_bulk_fetch_experiment_data_validation(self) -> None:
        exp_1 = get_experiment()
        exp_2 = get_experiment()
        exp_2.new_trial()
        m = Metric(name="test")
        with self.assertRaisesRegex(ValueError, "from the input experiment"):
            m.bulk_fetch_experiment_data(
                experiment=exp_1, trials=list(exp_2.trials.values()), metrics=[m]
            )

    def test_summary_dict(self) -> None:
        metric = Metric(name="m1", lower_is_better=False)
        self.assertDictEqual(
            metric.summary_dict,
            {
                "name": "m1",
                "type": "Metric",
                "lower_is_better": False,
            },
        )

        metric = TestMetric(name="m2", lower_is_better=True)
        self.assertDictEqual(
            metric.summary_dict,
            {
                "name": "m2",
                "type": "TestMetric",
                "lower_is_better": True,
            },
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, Dict, Tuple

import numpy as np
import pandas as pd
from ax.core.base_trial import BaseTrial
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.metric import Metric, MetricFetchE, MetricFetchResult
from ax.core.types import TParameterization, TParamValue
from ax.utils.common.result import Err, Ok
from ax.utils.stats.statstools import agresti_coull_sem


class FactorialMetric(Metric):
    """Metric for testing factorial designs assuming a main effects only
    logit model.
    """

    def __init__(
        self,
        name: str,
        coefficients: Dict[str, Dict[TParamValue, float]],
        batch_size: int = 10000,
        noise_var: float = 0.0,
    ) -> None:
        """
        Args:
            name: name of the metric.
            coefficients: a dictionary mapping
                factors to levels to main effects.
            batch_size: the sample size for one batch, distributed
                between arms proportionally to the design.
            noise_var: used in calculating the probability of
                each arm.
        """
        super(FactorialMetric, self).__init__(name)

        self.coefficients = coefficients
        self.batch_size = batch_size
        self.noise_var = noise_var

    @classmethod
    def is_available_while_running(cls) -> bool:
        # This metric does not require a trial to complete to fetch its
        # data, since there is no actual "data" to be fetched –– its
        # fabricated from parameterizations.
        return True

    def fetch_trial_data(self, trial: BaseTrial, **kwargs: Any) -> MetricFetchResult:
        try:
            if not isinstance(trial, BatchTrial):
                raise ValueError(
                    "Factorial metric can only fetch data for batch trials."
                )
            if not trial.status.expecting_data:
                raise ValueError("Can only fetch data if trial is expecting data.")

            data = []
            normalized_arm_weights = trial.normalized_arm_weights()
            for name, arm in trial.arms_by_name.items():
                weight = normalized_arm_weights[arm]
                mean, sem = evaluation_function(
                    parameterization=arm.parameters,
                    weight=weight,
                    coefficients=self.coefficients,
                    batch_size=self.batch_size,
                    noise_var=self.noise_var,
                )
                n = np.random.binomial(self.batch_size, weight)
                data.append(
                    {
                        "arm_name": name,
                        "metric_name": self.name,
                        "mean": mean,
                        "sem": sem,
                        "trial_index": trial.index,
                        "n": n,
                        "frac_nonnull": mean,
                    }
                )
            return Ok(value=Data(df=pd.DataFrame(data)))

        except Exception as e:
            return Err(
                MetricFetchE(message=f"Failed to fetch {self.name}", exception=e)
            )


def evaluation_function(
    parameterization: TParameterization,
    coefficients: Dict[str, Dict[TParamValue, float]],
    weight: float = 1.0,
    batch_size: int = 10000,
    noise_var: float = 0.0,
) -> Tuple[float, float]:
    probability = _parameterization_probability(
        parameterization=parameterization,
        coefficients=coefficients,
        noise_var=noise_var,
    )
    plays = np.random.binomial(batch_size, weight)
    successes = np.random.binomial(plays, probability)
    mean = float(successes) / plays
    sem = agresti_coull_sem(successes, plays)
    assert isinstance(sem, float)
    return mean, sem


def _parameterization_probability(
    parameterization: TParameterization,
    coefficients: Dict[str, Dict[TParamValue, float]],
    noise_var: float = 0.0,
) -> float:
    z = 0.0
    for factor, level in parameterization.items():
        if factor not in coefficients.keys():
            raise ValueError("{} not in supplied coefficients".format(factor))
        if level not in coefficients[factor].keys():
            raise ValueError("{} not a valid level of {}".format(level, factor))
        z += coefficients[factor][level]
    z += np.sqrt(noise_var) * np.random.randn()
    return np.exp(z) / (1 + np.exp(z))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import dataclasses
import datetime
import enum
from collections import OrderedDict
from inspect import isclass
from typing import Any, Callable, Dict, Type

import numpy as np
import pandas as pd
import torch
from ax.exceptions.storage import JSONEncodeError
from ax.storage.json_store.encoders import tensor_to_dict
from ax.storage.json_store.registry import (
    CORE_CLASS_ENCODER_REGISTRY,
    CORE_ENCODER_REGISTRY,
)
from ax.utils.common.serialization import _is_named_tuple
from ax.utils.common.typeutils_nonnative import numpy_type_to_python_type
from ax.utils.common.typeutils_torch import torch_type_to_str


# pyre-fixme[3]: Return annotation cannot be `Any`.
def object_to_json(  # noqa C901
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    obj: Any,
    # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    encoder_registry: Dict[
        Type, Callable[[Any], Dict[str, Any]]
    ] = CORE_ENCODER_REGISTRY,
    # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    class_encoder_registry: Dict[
        Type, Callable[[Any], Dict[str, Any]]
    ] = CORE_CLASS_ENCODER_REGISTRY,
) -> Any:
    """Convert an Ax object to a JSON-serializable dictionary.

    The root node passed to this function should always be an instance of a
    core Ax class or a JSON-compatible python builtin. The sub-fields of the
    input will then be recursively passed to this function.

    e.g. if we pass an instance of Experiment, we will first fall through
    to the line `object_dict = ENCODER_REGISTRY[_type](object)`, which
    will convert the Experiment to a (shallow) dictionary, where search
    subfield remains "unconverted", i.e.:
    {"name": <name: string>, "search_space": <search space: SearchSpace>}.
    We then pass each item of the dictionary back into this function to
    recursively convert the entire object.
    """
    obj = numpy_type_to_python_type(obj)
    _type = type(obj)

    # Type[MyClass] encoding (encoding of classes, not instances)
    if isclass(obj):
        for class_type in class_encoder_registry:
            if issubclass(obj, class_type):

                obj_dict = class_encoder_registry[class_type](obj)
                return {
                    k: object_to_json(
                        v,
                        encoder_registry=encoder_registry,
                        class_encoder_registry=class_encoder_registry,
                    )
                    for k, v in obj_dict.items()
                }

        raise ValueError(
            f"{obj} is a class. Add it to the CLASS_ENCODER_REGISTRY "
            "(and remove it from the ENCODER_REGISTRY if needed)."
        )

    if _type in encoder_registry:
        obj_dict = encoder_registry[_type](obj)
        return {
            k: object_to_json(
                v,
                encoder_registry=encoder_registry,
                class_encoder_registry=class_encoder_registry,
            )
            for k, v in obj_dict.items()
        }

    # Python built-in types + `typing` module types
    if _type in (str, int, float, bool, type(None)):
        return obj
    elif _type is list:
        return [
            object_to_json(
                x,
                encoder_registry=encoder_registry,
                class_encoder_registry=class_encoder_registry,
            )
            for x in obj
        ]
    elif _type is tuple:
        return tuple(
            object_to_json(
                x,
                encoder_registry=encoder_registry,
                class_encoder_registry=class_encoder_registry,
            )
            for x in obj
        )
    elif _type is dict:
        return {
            k: object_to_json(
                v,
                encoder_registry=encoder_registry,
                class_encoder_registry=class_encoder_registry,
            )
            for k, v in obj.items()
        }
    elif _is_named_tuple(obj):
        return {
            "__type": _type.__name__,
            **{
                k: object_to_json(
                    v,
                    encoder_registry=encoder_registry,
                    class_encoder_registry=class_encoder_registry,
                )
                for k, v in obj._asdict().items()
            },
        }
    elif dataclasses.is_dataclass(obj):
        return {
            "__type": _type.__name__,
            **{
                k: object_to_json(
                    v,
                    encoder_registry=encoder_registry,
                    class_encoder_registry=class_encoder_registry,
                )
                for k, v in obj.__dict__.items()
            },
        }

    # Types from libraries, commonly used in Ax (e.g., numpy, pandas, torch)
    elif _type is OrderedDict:
        return {
            "__type": _type.__name__,
            "value": [
                (
                    k,
                    object_to_json(
                        v,
                        encoder_registry=encoder_registry,
                        class_encoder_registry=class_encoder_registry,
                    ),
                )
                for k, v in obj.items()
            ],
        }
    elif _type is datetime.datetime:
        return {
            "__type": _type.__name__,
            "value": datetime.datetime.strftime(obj, "%Y-%m-%d %H:%M:%S.%f"),
        }
    elif _type is pd.DataFrame:
        return {"__type": _type.__name__, "value": obj.to_json()}
    elif issubclass(_type, enum.Enum):
        return {"__type": _type.__name__, "name": obj.name}
    elif _type is np.ndarray or issubclass(_type, np.ndarray):
        return {"__type": _type.__name__, "value": obj.tolist()}
    elif _type is set:
        return {"__type": _type.__name__, "value": list(obj)}
    elif _type is torch.Tensor:
        return tensor_to_dict(obj=obj)
    elif _type.__module__ == "torch":
        # Torch does not support saving to string, so save to buffer first
        return {"__type": f"torch_{_type.__name__}", "value": torch_type_to_str(obj)}

    err = (
        f"Object {obj} passed to `object_to_json` (of type {_type}, module: "
        f"{_type.__module__}) is not registered with a corresponding encoder "
        "in ENCODER_REGISTRY."
    )
    raise JSONEncodeError(err)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import math
from typing import List, Optional, Set, TYPE_CHECKING

from ax.core.observation import Observation, ObservationFeatures
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.modelbridge.transforms.base import Transform
from ax.models.types import TConfig

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import modelbridge as modelbridge_module  # noqa F401


class Log(Transform):
    """Apply log base 10 to a float RangeParameter domain.

    Transform is done in-place.
    """

    def __init__(
        self,
        search_space: Optional[SearchSpace] = None,
        observations: Optional[List[Observation]] = None,
        modelbridge: Optional["modelbridge_module.base.ModelBridge"] = None,
        config: Optional[TConfig] = None,
    ) -> None:
        assert search_space is not None, "Log requires search space"
        # Identify parameters that should be transformed
        self.transform_parameters: Set[str] = {
            p_name
            for p_name, p in search_space.parameters.items()
            if isinstance(p, RangeParameter)
            and p.parameter_type == ParameterType.FLOAT
            and p.log_scale is True
        }

    def transform_observation_features(
        self, observation_features: List[ObservationFeatures]
    ) -> List[ObservationFeatures]:
        for obsf in observation_features:
            for p_name in self.transform_parameters:
                if p_name in obsf.parameters:
                    # pyre: param is declared to have type `float` but is used
                    # pyre-fixme[9]: as type `Optional[typing.Union[bool, float, str]]`.
                    param: float = obsf.parameters[p_name]
                    obsf.parameters[p_name] = math.log10(param)
        return observation_features

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        for p_name, p in search_space.parameters.items():
            if p_name in self.transform_parameters and isinstance(p, RangeParameter):
                p.set_log_scale(False).update_range(
                    lower=math.log10(p.lower), upper=math.log10(p.upper)
                )
                if p.target_value is not None:
                    p._target_value = math.log10(p.target_value)  # pyre-ignore [6]
        return search_space

    def untransform_observation_features(
        self, observation_features: List[ObservationFeatures]
    ) -> List[ObservationFeatures]:
        for obsf in observation_features:
            for p_name in self.transform_parameters:
                if p_name in obsf.parameters:
                    # pyre: param is declared to have type `float` but is used
                    # pyre-fixme[9]: as type `Optional[typing.Union[bool, float, str]]`.
                    param: float = obsf.parameters[p_name]
                    obsf.parameters[p_name] = math.pow(10, param)
        return observation_features



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from datetime import datetime

import numpy as np
import pandas as pd
from ax.utils.common.equality import (
    dataframe_equals,
    datetime_equals,
    equality_typechecker,
    object_attribute_dicts_find_unequal_fields,
    same_elements,
)
from ax.utils.common.testutils import TestCase


class EqualityTest(TestCase):
    def test_EqualityTypechecker(self) -> None:
        @equality_typechecker
        # pyre-fixme[3]: Return type must be annotated.
        # pyre-fixme[2]: Parameter must be annotated.
        def eq(x, y):
            return x == y

        self.assertFalse(eq(5, 5.0))
        self.assertTrue(eq(5, 5))

    def test_ListsEquals(self) -> None:
        self.assertFalse(same_elements([0], [0, 1]))
        self.assertFalse(same_elements([1, 0], [0, 2]))
        self.assertFalse(same_elements([1, 1], [1, 2]))
        self.assertFalse(same_elements([1, 2], [1, 1]))
        self.assertFalse(same_elements([1, 1, 2], [1, 2, 2]))
        self.assertTrue(same_elements([1, 0], [0, 1]))

    def test_ListsEquals_floats(self) -> None:
        self.assertTrue(same_elements([0.0], [0.000000000000001]))
        self.assertTrue(same_elements([float("nan")], [float("nan")]))

    def test_DatetimeEquals(self) -> None:
        now = datetime.now()
        self.assertTrue(datetime_equals(None, None))
        self.assertFalse(datetime_equals(None, now))
        self.assertTrue(datetime_equals(now, now))

    def test_DataframeEquals(self) -> None:
        pd1 = pd.DataFrame.from_records([{"x": 100, "y": 200}])
        pd2 = pd.DataFrame.from_records([{"y": 200, "x": 100}])
        pd3 = pd.DataFrame.from_records([{"x": 100, "y": 300}])

        self.assertTrue(dataframe_equals(pd.DataFrame(), pd.DataFrame()))
        self.assertTrue(dataframe_equals(pd1, pd2))
        self.assertFalse(dataframe_equals(pd1, pd3))

    def test_numpy_equals(self) -> None:
        # Simple check.
        np_0 = {"cov": np.array([[0.1, 0.0], [0.0, 0.1]])}
        np_1 = {"cov": np.array([[0.1, 0.0], [0.0, 0.1]])}
        self.assertEqual(
            object_attribute_dicts_find_unequal_fields(np_0, np_1), ({}, {})
        )
        # Unequal.
        np_1 = {"cov": np.array([[0.1, 0.0], [0.1, 0.1]])}
        self.assertEqual(
            object_attribute_dicts_find_unequal_fields(np_0, np_1),
            ({}, {"cov": (np_0["cov"], np_1["cov"])}),
        )
        # With NaNs.
        np_1 = {"cov": np.array([[0.1, float("nan")], [float("nan"), 0.1]])}
        self.assertEqual(
            object_attribute_dicts_find_unequal_fields(np_0, np_1),
            ({}, {"cov": (np_0["cov"], np_1["cov"])}),
        )
        np_0 = {"cov": np.array([[0.1, float("nan")], [float("nan"), 0.1]])}
        self.assertEqual(
            object_attribute_dicts_find_unequal_fields(np_0, np_1), ({}, {})
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from enum import Enum
from typing import cast, Dict, List, Optional, Tuple, Type, Union

import numpy as np
from ax.core.data import Data
from ax.core.map_data import MapData
from ax.core.types import (
    TEvaluationOutcome,
    TMapTrialEvaluation,
    TTrialEvaluation,
    validate_evaluation_outcome,
)
from ax.exceptions.core import UserInputError
from ax.utils.common.typeutils_nonnative import numpy_type_to_python_type


# -------------------- Data formatting utils. ---------------------


class DataType(Enum):
    DATA = 1
    MAP_DATA = 3


DATA_TYPE_LOOKUP: Dict[DataType, Type[Data]] = {
    DataType.DATA: Data,
    DataType.MAP_DATA: MapData,
}


def raw_data_to_evaluation(
    raw_data: TEvaluationOutcome,
    metric_names: List[str],
) -> TEvaluationOutcome:
    """Format the trial evaluation data to a standard `TTrialEvaluation`
    (mapping from metric names to a tuple of mean and SEM) representation, or
    to a TMapTrialEvaluation.

    Note: this function expects raw_data to be data for a `Trial`, not a
    `BatchedTrial`.
    """
    if isinstance(raw_data, dict):
        if any(isinstance(x, dict) for x in raw_data.values()):
            raise UserInputError("Raw data is expected to be just for one arm.")
        for metric_name, dat in raw_data.items():
            if not isinstance(dat, tuple):
                if not isinstance(dat, (float, int)):
                    raise UserInputError(
                        "Raw data for an arm is expected to either be a tuple of "
                        "numerical mean and SEM or just a numerical mean. "
                        f"Got: {dat} for metric '{metric_name}'."
                    )
                raw_data[metric_name] = (float(dat), None)
        return raw_data
    try:
        validate_evaluation_outcome(outcome=raw_data)
    except Exception as e:
        raise UserInputError(
            "Raw data does not conform to the expected structure. For simple "
            "evaluations of one or more metrics, `raw_data` is expected to be "
            "a dictionary of the form `{metric_name -> mean}` or `{metric_name "
            "-> (mean, SEM)}`. For fidelity or mapping (e.g., early stopping) "
            "evaluation, the expected format is `[(fidelities, {metric_name -> "
            "(mean, SEM)})]` or `[({mapping_key, mapping_value}, {metric_name -> "
            "(mean, SEM)})]`."
            f"Received {raw_data=}. Original validation error: {e}."
        )
    if isinstance(raw_data, list):
        validate_evaluation_outcome(raw_data)
        return raw_data
    elif len(metric_names) > 1:
        raise UserInputError(
            "Raw data must be a dictionary of metric names to mean "
            "for experiments with multiple metrics attached. "
            f"Got {raw_data=} for {metric_names=}."
        )
    elif isinstance(raw_data, tuple):
        return {metric_names[0]: raw_data}
    elif isinstance(raw_data, (float, int)):
        return {metric_names[0]: (raw_data, None)}
    elif isinstance(raw_data, (np.float32, np.float64, np.int32, np.int64)):
        return {metric_names[0]: (numpy_type_to_python_type(raw_data), None)}
    else:
        raise UserInputError(
            "Raw data has an invalid type. The data must either be in the form "
            "of a dictionary of metric names to mean, sem tuples, "
            "or a single mean, sem tuple, or a single mean."
        )


def data_and_evaluations_from_raw_data(
    raw_data: Dict[str, TEvaluationOutcome],
    metric_names: List[str],
    trial_index: int,
    sample_sizes: Dict[str, int],
    data_type: DataType,
    start_time: Optional[Union[int, str]] = None,
    end_time: Optional[Union[int, str]] = None,
) -> Tuple[Dict[str, TEvaluationOutcome], Data]:
    """Transforms evaluations into Ax Data.

    Each evaluation is either a trial evaluation: {metric_name -> (mean, SEM)}
    or a fidelity trial evaluation for multi-fidelity optimizations:
    [(fidelities, {metric_name -> (mean, SEM)})].

    Args:
        raw_data: Mapping from arm name to raw_data.
        metric_names: Names of metrics used to transform raw data to evaluations.
        trial_index: Index of the trial, for which the evaluations are.
        sample_sizes: Number of samples collected for each arm, may be empty
            if unavailable.
        start_time: Optional start time of run of the trial that produced this
            data, in milliseconds or iso format.  Milliseconds will eventually be
            converted to iso format because iso format automatically works with the
            pandas column type `Timestamp`.
        end_time: Optional end time of run of the trial that produced this
            data, in milliseconds or iso format.  Milliseconds will eventually be
            converted to iso format because iso format automatically works with the
            pandas column type `Timestamp`.
    """
    evaluations = {
        arm_name: raw_data_to_evaluation(
            raw_data=raw_data[arm_name],
            metric_names=metric_names,
        )
        for arm_name in raw_data
    }
    if all(isinstance(evaluations[x], dict) for x in evaluations.keys()):
        if data_type is DataType.MAP_DATA:
            raise UserInputError(
                "The format of the `raw_data` is not compatible with `MapData`. "
                "Possible cause: Did you set default data type to `MapData`, e.g., "
                "for early stopping, but forgot to provide the `raw_data` "
                "in the form of `[(fidelities, {metric_name -> (mean, SEM)})]` or "
                "`[({mapping_key, mapping_value}, {metric_name -> (mean, SEM)})]`? "
                f"Received: {raw_data=}"
            )
        # All evaluations are no-fidelity evaluations.
        data = Data.from_evaluations(
            evaluations=cast(Dict[str, TTrialEvaluation], evaluations),
            trial_index=trial_index,
            sample_sizes=sample_sizes,
            start_time=start_time,
            end_time=end_time,
        )
    elif all(isinstance(evaluations[x], list) for x in evaluations.keys()):
        if data_type is DataType.DATA:
            raise UserInputError(
                "The format of the `raw_data` is not compatible with `Data`. "
                "Possible cause: Did you provide data for multi-fidelity evaluations, "
                "e.g., for early stopping, but forgot to set the default data type "
                f"to `MapData`? Received: {raw_data=}"
            )
        # All evaluations are map evaluations.
        data = MapData.from_map_evaluations(
            evaluations=cast(Dict[str, TMapTrialEvaluation], evaluations),
            trial_index=trial_index,
        )
    else:
        raise ValueError(
            "Evaluations included a mixture of no-fidelity and with-fidelity "
            "evaluations, which is not currently supported."
        )
    return evaluations, data



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from abc import ABC, abstractmethod
from math import sqrt
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial
from ax.core.batch_trial import BatchTrial
from ax.core.runner import Runner
from ax.core.trial import Trial

from ax.utils.common.typeutils import checked_cast
from torch import Tensor


class BenchmarkRunner(Runner, ABC):

    @property
    @abstractmethod
    def outcome_names(self) -> List[str]:
        """The names of the outcomes of the problem (in the order of the outcomes)."""
        pass  # pragma: no cover

    def get_Y_true(self, arm: Arm) -> Tensor:
        """Function returning the ground truth values for a given arm. The
        synthetic noise is added as part of the Runner's `run()` method.
        For problems that do not have a ground truth, the Runner must
        implement the `get_Y_Ystd()` method instead."""
        raise NotImplementedError(
            "Must implement method `get_Y_true()` for Runner "
            f"{self.__class__.__name__} as it does not implement a "
            "`get_Y_Ystd()` method."
        )

    def get_noise_stds(self) -> Union[None, float, Dict[str, float]]:
        """Function returning the standard errors for the synthetic noise
        to be applied to the observed values. For problems that do not have
        a ground truth, the Runner must implement the `get_Y_Ystd()` method
        instead."""
        raise NotImplementedError(
            "Must implement method `get_Y_Ystd()` for Runner "
            f"{self.__class__.__name__} as it does not implement a "
            "`get_noise_stds()` method."
        )

    def get_Y_Ystd(self, arm: Arm) -> Tuple[Tensor, Optional[Tensor]]:
        """Function returning the observed values and their standard errors
        for a given arm. This function is unused for problems that have a
        ground truth (in this case `get_Y_true()` is used), and is required
        for problems that do not have a ground truth."""
        raise NotImplementedError(
            "Must implement method `get_Y_Ystd()` for Runner "
            f"{self.__class__.__name__} as it does not implement a "
            "`get_Y_true()` method."
        )

    def run(self, trial: BaseTrial) -> Dict[str, Any]:
        """Run the trial by evaluating its parameterization(s).

        Args:
            trial: The trial to evaluate.

        Returns:
            A dictionary with the following keys:
                - Ys: A dict mapping arm names to lists of corresponding outcomes,
                    where the order of the outcomes is the same as in `outcome_names`.
                - Ystds: A dict mapping arm names to lists of corresponding outcome
                    noise standard deviations (possibly nan if the noise level is
                    unobserved), where the order of the outcomes is the same as in
                    `outcome_names`.
                - Ys_true: A dict mapping arm names to lists of corresponding ground
                    truth outcomes, where the order of the outcomes is the same as
                    in `outcome_names`. If the benchmark problem does not provide a
                    ground truth, this key will not be present in the dict returned
                    by this function.
                - "outcome_names": A list of metric names.
        """
        Ys, Ys_true, Ystds = {}, {}, {}
        noise_stds = self.get_noise_stds()

        if noise_stds is not None:
            # extract arm weights to adjust noise levels accordingly
            if isinstance(trial, BatchTrial):
                # normalize arm weights (we assume that the noise level is defined)
                # w.r.t. to a single arm allocated all of the sample budget
                nlzd_arm_weights = {
                    arm: weight / sum(trial.arm_weights.values())
                    for arm, weight in trial.arm_weights.items()
                }
            else:
                nlzd_arm_weights = {checked_cast(Trial, trial).arm: 1.0}
            # generate a tensor of noise levels that we'll reuse below
            if isinstance(noise_stds, float):
                noise_stds_tsr = torch.full(
                    (len(self.outcome_names),),
                    noise_stds,
                    dtype=torch.double,
                )
            else:
                noise_stds_tsr = torch.tensor(
                    [noise_stds[metric_name] for metric_name in self.outcome_names],
                    dtype=torch.double,
                )

        for arm in trial.arms:
            try:
                # Case where we do have a ground truth
                Y_true = self.get_Y_true(arm)
                Ys_true[arm.name] = Y_true.tolist()
                if noise_stds is None:
                    # No noise, so just return the true outcome.
                    Ystds[arm.name] = [0.0] * len(Y_true)
                    Ys[arm.name] = Y_true.tolist()
                else:
                    # We can scale the noise std by the inverse of the relative sample
                    # budget allocation to each arm. This works b/c (i) we assume that
                    # observations per unit sample budget are i.i.d. and (ii) the
                    # normalized weights sum to one.
                    std = noise_stds_tsr.to(Y_true) / sqrt(nlzd_arm_weights[arm])
                    Ystds[arm.name] = std.tolist()
                    Ys[arm.name] = (Y_true + std * torch.randn_like(Y_true)).tolist()
            except NotImplementedError:
                # Case where we don't have a ground truth.
                Y, Ystd = self.get_Y_Ystd(arm)
                Ys[arm.name] = Y.tolist()
                Ystds[arm.name] = Ystd.tolist() if Ystd is not None else None

        run_metadata = {
            "Ys": Ys,
            "Ystds": Ystds,
            "outcome_names": self.outcome_names,
        }
        if Ys_true:  # only add key if we actually have a ground truth
            run_metadata["Ys_true"] = Ys_true
        return run_metadata



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import List, Optional, TYPE_CHECKING

from ax.core.observation import Observation, ObservationData
from ax.core.search_space import SearchSpace
from ax.modelbridge.transforms.base import Transform
from ax.modelbridge.transforms.utils import match_ci_width_truncated
from ax.models.types import TConfig
from ax.utils.common.logger import get_logger
from scipy.stats import norm


if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax.modelbridge import base as base_modelbridge  # noqa F401


logger: Logger = get_logger(__name__)


# TODO(jej): Add OptimizationConfig validation - can't transform outcome constraints.
class InverseGaussianCdfY(Transform):
    """Apply inverse CDF transform to Y.

    This means that we model uniform distributions as gaussian-distributed.
    """

    def __init__(
        self,
        search_space: Optional[SearchSpace] = None,
        observations: Optional[List[Observation]] = None,
        modelbridge: Optional["base_modelbridge.ModelBridge"] = None,
        config: Optional[TConfig] = None,
    ) -> None:
        # pyre-fixme[4]: Attribute must be annotated.
        self.dist = norm(loc=0, scale=1)

    def _transform_observation_data(
        self,
        observation_data: List[ObservationData],
    ) -> List[ObservationData]:
        """Map to inverse Gaussian CDF in place."""
        # TODO (jej): Transform covariances.
        for obsd in observation_data:
            for idx, _ in enumerate(obsd.metric_names):
                mean = float(obsd.means[idx])
                # Error on out-of-domain values.
                if mean <= 0.0 or mean >= 1.0:
                    raise ValueError(
                        f"Inverse CDF cannot transform value: {mean} outside (0, 1)"
                    )
                var = float(obsd.covariance[idx, idx])
                transformed_mean, transformed_var = match_ci_width_truncated(
                    mean, var, self._map, lower_bound=0.0, upper_bound=1.0
                )
                obsd.means[idx] = transformed_mean
                obsd.covariance[idx, idx] = transformed_var
        return observation_data

    def _map(self, val: float) -> float:
        mapped_val = self.dist.ppf(val)
        return mapped_val



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, Dict, List, Optional

import torch
from ax.models.torch.botorch import BotorchModel
from botorch.fit import fit_gpytorch_mll
from botorch.models.contextual_multioutput import LCEMGP
from botorch.models.model_list_gp_regression import ModelListGP
from gpytorch.mlls.sum_marginal_log_likelihood import SumMarginalLogLikelihood
from torch import Tensor


MIN_OBSERVED_NOISE_LEVEL = 1e-7


class LCEMBO(BotorchModel):
    r"""Does Bayesian optimization with LCE-M GP."""

    def __init__(
        self,
        context_cat_feature: Optional[Tensor] = None,
        context_emb_feature: Optional[Tensor] = None,
        embs_dim_list: Optional[List[int]] = None,
    ) -> None:
        self.context_cat_feature = context_cat_feature
        self.context_emb_feature = context_emb_feature
        self.embs_dim_list = embs_dim_list
        super().__init__(model_constructor=self.get_and_fit_model)

    def get_and_fit_model(
        self,
        Xs: List[Tensor],
        Ys: List[Tensor],
        Yvars: List[Tensor],
        task_features: List[int],
        fidelity_features: List[int],
        metric_names: List[str],
        state_dict: Optional[Dict[str, Tensor]] = None,
        fidelity_model_id: Optional[int] = None,
        **kwargs: Any,
    ) -> ModelListGP:
        """Get a fitted multi-task contextual GP model for each outcome.
        Args:
            Xs: List of X data, one tensor per outcome.
            Ys: List of Y data, one tensor per outcome.
            Yvars:List of Noise variance of Yvar data, one tensor per outcome.
            task_features: List of columns of X that are tasks.
        Returns: ModeListGP that each model is a fitted LCEM GP model.
        """

        if len(task_features) == 1:
            task_feature = task_features[0]
        elif len(task_features) > 1:
            raise NotImplementedError(
                f"LCEMBO only supports 1 task feature (got {task_features})"
            )
        else:
            raise ValueError("LCEMBO requires context input as task features")

        models = []
        for i, X in enumerate(Xs):
            # validate input Yvars
            Yvar = Yvars[i].clamp_min_(MIN_OBSERVED_NOISE_LEVEL)
            is_nan = torch.isnan(Yvar)
            all_nan_Yvar = torch.all(is_nan)
            all_tasks, _, _ = LCEMGP.get_all_tasks(train_X=X, task_feature=task_feature)
            gp_m = LCEMGP(
                train_X=X,
                train_Y=Ys[i],
                train_Yvar=None if all_nan_Yvar else Yvar,
                task_feature=task_feature,
                context_cat_feature=self.context_cat_feature,
                context_emb_feature=self.context_emb_feature,
                embs_dim_list=self.embs_dim_list,
                # specify output tasks so that model.num_outputs = 1
                # since the model only models a single outcome.
                output_tasks=all_tasks[:1],
            )

            models.append(gp_m)
        # Use a ModelListGP
        model = ModelListGP(*models)
        model.to(Xs[0])
        mll = SumMarginalLogLikelihood(model.likelihood, model)
        fit_gpytorch_mll(mll)
        return model



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from collections import defaultdict
from typing import Any, Callable, Dict, Iterable, Optional, Set

import numpy as np
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.runner import Runner
from ax.utils.testing.backend_simulator import BackendSimulator


class SimulatedBackendRunner(Runner):
    """Class for a runner that works with the BackendSimulator."""

    def __init__(
        self,
        simulator: BackendSimulator,
        sample_runtime_func: Optional[Callable[[BaseTrial], float]] = None,
    ) -> None:
        """Runner for a BackendSimulator.

        Args:
            simulator: The backend simulator.
            sample_runtime_func: A Callable that samples a runtime given a trial.
        """
        self.simulator = simulator
        if sample_runtime_func is None:
            sample_runtime_func = sample_runtime_unif
        self.sample_runtime_func: Callable[[BaseTrial], float] = sample_runtime_func

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> Dict[TrialStatus, Set[int]]:
        """Poll trial status from the ``BackendSimulator``. NOTE: The ``Scheduler``
        currently marks trials as running when they are created, but some of these
        trials may actually be in queued on the ``BackendSimulator``.

        Returns:
            A Dict mapping statuses to sets of trial indices.
        """
        self.simulator.update()
        trial_status = defaultdict(set)
        for trial in trials:
            t_index = trial.index
            status = self.simulator.lookup_trial_index_status(t_index)
            trial_status[status].add(t_index)
        return dict(trial_status)

    def run(self, trial: BaseTrial) -> Dict[str, Any]:
        """Start a trial on the BackendSimulator.

        Args:
            trial: Trial to deploy via the runner.

        Returns:
            Dict containing the sampled runtime of the trial.
        """
        runtime = self.sample_runtime_func(trial)
        self.simulator.run_trial(trial_index=trial.index, runtime=runtime)
        return {"runtime": runtime}

    def stop(self, trial: BaseTrial, reason: Optional[str] = None) -> Dict[str, Any]:
        """Stop a trial on the BackendSimulator.

        Args:
            trial: Trial to stop on the simulator.
            reason: A message containing information why the trial is to be stopped.

        Returns:
            A dictionary containing a single key "reason" that maps to the reason
            passed to the function. If no reason was given, returns an empty dictionary.
        """
        self.simulator.stop_trial(trial.index)
        return {"reason": reason} if reason else {}


def sample_runtime_unif(trial: BaseTrial, low: float = 1.0, high: float = 5.0) -> float:
    """Return a uniform runtime in [low, high]

    Args:
        trial: Trial for which to sample runtime.
        low: Lower bound of uniform runtime distribution.
        high: Upper bound of uniform runtime distribution.

    Returns:
        A float representing the simulated trial runtime.
    """
    return np.random.uniform(low, high)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import warnings
from typing import Any, Dict, Iterable, List, Optional, Set, Union

import torch
from ax.benchmark.runners.base import BenchmarkRunner
from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.observation import ObservationFeatures
from ax.core.search_space import SearchSpace
from ax.modelbridge.torch import TorchModelBridge
from ax.utils.common.serialization import TClassDecoderRegistry, TDecoderRegistry
from botorch.utils.datasets import SupervisedDataset
from torch import Tensor


class SurrogateRunner(BenchmarkRunner):
    def __init__(
        self,
        name: str,
        surrogate: TorchModelBridge,
        datasets: List[SupervisedDataset],
        search_space: SearchSpace,
        outcome_names: List[str],
        noise_stds: Union[float, Dict[str, float]] = 0.0,
    ) -> None:
        """Runner for surrogate benchmark problems.

        Args:
            name: The name of the runner.
            surrogate: The modular BoTorch model `Surrogate` to use for
                generating observations.
            search_space: The search space of the problem (used for
                parameter transforms).
            datasets: The data sets used to fit the surrogate model.
            outcome_names: The names of the outcomes of the Surrogate.
            noise_stds: Noise standard deviations to add to the surrogate output(s).
                If a single float is provided, noise with that standard deviation
                is added to all outputs. Alternatively, a dictionary mapping outcome
                names to noise standard deviations can be provided to specify different
                noise levels for different outputs.
        """
        self.name = name
        self.surrogate = surrogate
        self._outcome_names = outcome_names
        self.datasets = datasets
        self.search_space = search_space
        self.noise_stds = noise_stds
        self.statuses: Dict[int, TrialStatus] = {}

    @property
    def outcome_names(self) -> List[str]:
        return self._outcome_names

    def get_noise_stds(self) -> Union[None, float, Dict[str, float]]:
        return self.noise_stds

    def get_Y_true(self, arm: Arm) -> Tensor:
        # We're ignoring the uncertainty predictions of the surrogate model here and
        # use the mean predictions as the outcomes (before potentially adding noise)
        means, _ = self.surrogate.predict(
            observation_features=[ObservationFeatures(arm.parameters)]
        )
        means = [means[name][0] for name in self.outcome_names]
        return torch.tensor(
            means,
            device=self.surrogate.device,
            dtype=self.surrogate.dtype,
        )

    def run(self, trial: BaseTrial) -> Dict[str, Any]:
        """Run the trial by evaluating its parameterization(s) on the surrogate model.

        Note: This also sets the status of the trial to COMPLETED.

        Args:
            trial: The trial to evaluate.

        Returns:
            A dictionary with the following keys:
                - outcome_names: The names of the metrics being evaluated.
                - Ys: A dict mapping arm names to lists of corresponding outcomes,
                    where the order of the outcomes is the same as in `outcome_names`.
                - Ystds: A dict mapping arm names to lists of corresponding outcome
                    noise standard deviations (possibly nan if the noise level is
                    unobserved), where the order of the outcomes is the same as in
                    `outcome_names`.
                - Ys_true: A dict mapping arm names to lists of corresponding ground
                    truth outcomes, where the order of the outcomes is the same as
                    in `outcome_names`.
        """
        self.statuses[trial.index] = TrialStatus.COMPLETED
        run_metadata = super().run(trial=trial)
        run_metadata["outcome_names"] = self.outcome_names
        return run_metadata

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> Dict[TrialStatus, Set[int]]:
        return {TrialStatus.COMPLETED: {t.index for t in trials}}

    @classmethod
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def serialize_init_args(cls, obj: Any) -> Dict[str, Any]:
        """Serialize the properties needed to initialize the runner.
        Used for storage.

        WARNING: Because of issues with consistently saving and loading BoTorch and
        GPyTorch modules the SurrogateRunner cannot be serialized at this time.
        At load time the runner will be replaced with a SyntheticRunner.
        """
        warnings.warn(
            "Because of issues with consistently saving and loading BoTorch and "
            f"GPyTorch modules, {cls.__name__} cannot be serialized at this time. "
            "At load time the runner will be replaced with a SyntheticRunner.",
            stacklevel=3,
        )
        return {}

    @classmethod
    def deserialize_init_args(
        cls,
        args: Dict[str, Any],
        decoder_registry: Optional[TDecoderRegistry] = None,
        class_decoder_registry: Optional[TClassDecoderRegistry] = None,
    ) -> Dict[str, Any]:
        return {}
